from .generic import *
from scipy import stats, ndimage as ndi
import cv2
import warnings
import pandas as pd

__all__ = ['multiscale_image_feature',
           'local_stat',
           'ratio_of_gaussian',
           'normalized_difference_of_gaussian',
           'gabor_kernel_bank',
           'basic_intensity_features']

def multiscale_image_feature(image,
                             sigmas=(0.5, 0.8, 1.5, 3.5),
                             rog_sigmas=(0.5, 5),
                             num_workers=None,
                             mode='reflect',
                             order='rc',
                             shapeindex=True,
                             sobel=True,
                             ridge=True,
                             rog=True,
                             cval=0,
                             ext_kernels={}):
    """
    Computes multiscale features for an image including Gaussian filtered images, shape index, Sobel edges, ridges, etc.

    :param image: Input image
    :type image: numpy.ndarray
    :param sigmas: The standard deviations for the Gaussian filter
    :type sigmas: tuple, optional
    :param rog_sigmas: The standard deviations for the ratio of Gaussians (RoG)
    :type rog_sigmas: tuple, optional
    :param num_workers: The number of worker threads for parallel processing
    :type num_workers: int, optional
    :param mode: How to handle the borders for the filters
    :type mode: str, optional
    :param order: The order for the axes (reversed by default)
    :type order: str, optional
    :param shapeindex: Whether to compute the shape index or not
    :type shapeindex: bool, optional
    :param sobel: Whether to compute the Sobel edges or not
    :type sobel: bool, optional
    :param ridge: Whether to compute the ridges or not
    :type ridge: bool, optional
    :param rog: Whether to compute the ratio of Gaussians or not
    :type rog: bool, optional
    :param cval: The constant value for padding the image before filtering
    :type cval: int, optional
    :param ext_kernels: Additional kernels to filter the image with
    :type ext_kernels: dict, optional
    :return: A dictionary of Gaussian filtered images and a dictionary of image features
    :rtype: dict, dict
    """

    from concurrent.futures import ThreadPoolExecutor

    if num_workers is None:
        num_workers = len(sigmas)

    with ThreadPoolExecutor(max_workers=num_workers) as ex:
        out_sigmas = list(
            ex.map(
                lambda s: _uniscale_image_feature(image, s,
                                                  mode=mode,
                                                  order=order,
                                                  shapeindex=shapeindex,
                                                  sobel=sobel,
                                                  ridge=ridge,
                                                  cval=cval,
                                                  ext_kernels=ext_kernels),
                sigmas,
            )
        )
    multiscale_gaussian = {}
    multiscale_features = {}
    for i, pair in enumerate(out_sigmas):
        gaussian_filtered, image_features = pair
        multiscale_gaussian[sigmas[i]] = gaussian_filtered
        multiscale_features[sigmas[i]] = image_features
    reformated_features = {'Sigma{}-{}'.format(sigma, k): v for sigma, dic in multiscale_features.items() for k, v in
                           dic.items()}

    # compute rog
    if rog:
        if rog_sigmas[0] in sigmas:
            g1 = multiscale_gaussian[rog_sigmas[0]]
        else:
            g1 = filters.gaussian(image, rog_sigmas[0])
        if rog_sigmas[1] in sigmas:
            g2 = multiscale_gaussian[rog_sigmas[1]]
        else:
            g2 = filters.gaussian(image, rog_sigmas[1])
        reformated_features['RoG'] = g2 / g1

    return multiscale_gaussian, reformated_features


def local_stat(target_images,
               x_coords,
               y_coords,
               selem=morphology.disk(2),
               as_dataframe=True):
    """
    Computes local statistics (min, max, std, mean) for a list of images at given coordinates.

    :param target_images: Dictionary of images
    :type target_images: dict
    :param x_coords: X-coordinates
    :type x_coords: numpy.ndarray
    :param y_coords: Y-coordinates
    :type y_coords: numpy.ndarray
    :param selem: The neighborhood expressed as a 2-D array of booleans
    :type selem: numpy.ndarray, optional
    :param as_dataframe: If True, return as pandas DataFrame, else return as numpy array
    :type as_dataframe: bool, optional
    :return: Local statistics for each image
    :rtype: pandas.DataFrame or numpy.ndarray
    """

    stacked_measures = []
    col_names = []

    if selem is None:
        for name, img in target_images.items():
            col_names.append('{}-intensity'.format(name))
            stacked_measures.append(img[x_coords, y_coords].reshape(-1, 1))
    else:
        x_mesh, y_mesh = coord2mesh(x_coords, y_coords, selem,
                                    target_images[list(target_images.keys())[0]].shape[0],
                                    target_images[list(target_images.keys())[0]].shape[1])

        for name, img in target_images.items():
            mesh_data = img[x_mesh, y_mesh]
            mesh_measures = np.array([img[x_coords, y_coords],
                                      np.min(mesh_data, axis=1),
                                      np.max(mesh_data, axis=1),
                                      np.std(mesh_data, axis=1),
                                      np.mean(mesh_data, axis=1)]).T
            col_names += ['{}_{}'.format(name, stat) for stat in ['intensity', 'min', 'max', 'std', 'mean']]
            stacked_measures.append(mesh_measures)
    if as_dataframe:
        return pd.DataFrame(np.hstack(stacked_measures), columns=col_names)
    else:
        return np.hstack(stacked_measures)


def _uniscale_image_feature(image,
                            sigma,
                            mode='reflect',
                            order='rc',
                            shapeindex=True,
                            sobel=True,
                            ridge=True,
                            ext_kernels={},
                            cval=0):
    """
    Computes uniscale features for an image including Gaussian filtered images, shape index, Sobel edges, ridges, etc.

    :param image: Input image
    :type image: numpy.ndarray
    :param sigma: The standard deviation for the Gaussian filter
    :type sigma: float
    :param mode: How to handle the borders for the filters
    :type mode: str, optional
    :param order: The order for the axes (reversed by default)
    :type order: str, optional
    :param shapeindex: Whether to compute the shape index or not
    :type shapeindex: bool, optional
    :param sobel: Whether to compute the Sobel edges or not
    :type sobel: bool, optional
    :param ridge: Whether to compute the ridges or not
    :type ridge: bool, optional
    :param ext_kernels: Additional kernels to filter the image with
    :type ext_kernels: dict, optional
    :param cval: The constant value for padding the image before filtering
    :type cval: int, optional
    :return: Gaussian filtered image and a dictionary of image features
    :rtype: numpy.ndarray, dict
    """

    from itertools import combinations_with_replacement
    from skimage.feature import hessian_matrix_eigvals

    gaussian_filtered = filters.gaussian(image,
                                         sigma=sigma,
                                         mode=mode,
                                         cval=cval)
    gradients = np.gradient(gaussian_filtered)
    axes = range(image.ndim)
    if order == 'rc':
        axes = reversed(axes)

    H_elems = [np.gradient(gradients[ax0], axis=ax1)
               for ax0, ax1 in combinations_with_replacement(axes, 2)]

    # correct for scale
    H_elems = [(sigma ** 2) * e for e in H_elems]
    l1, l2 = hessian_matrix_eigvals(H_elems)

    image_features = {'H_eig_1': l1, 'H_eig_2': l2}
    if shapeindex:
        # shape index
        shpid = (2.0 / np.pi) * np.arctan((l2 + l1) / (l2 - l1))
        shpid[np.isnan(shpid)] = 0
        image_features['shapeindex'] = shpid
    if sobel:
        # sobel edge
        image_features['sobel'] = filters.sobel(gaussian_filtered)

    if ridge:
        image_features['ridge'] = filters.sato(image, sigmas=(0.1, sigma), black_ridges=False, mode='constant')

    if len(ext_kernels) > 0:
        for name, kernel in ext_kernels.items():
            image_features[name] = cv2.filter2D(image,-1,kernel)
    return gaussian_filtered, image_features


def ratio_of_gaussian(img, s1=0.5, s2=5):
    g1 = filters.gaussian(img,sigma=s1)
    g2 = filters.gaussian(img,sigma=s2)
    return g1/g2


def hu_log10_transform(hu_moments):
    warnings.filterwarnings("ignore")
    abs_vals = np.abs(hu_moments)
    hu_log10 = np.copysign(np.log10(abs_vals), hu_moments)
    corrected = np.zeros(hu_moments.shape)
    corrected[np.isfinite(hu_log10)] = hu_log10[np.isfinite(hu_log10)]
    return corrected


def median_over_gaussian(data,windowsize=3):
    return (ndi.median_filter(data,size=windowsize)/filters.gaussian(data,sigma=windowsize,preserve_range=True))


def normalized_difference_of_gaussian(data, s1=0, s2=10):
    g1 = filters.gaussian(data, sigma=s1)
    g2 = filters.gaussian(data, sigma=s2)
    return (g1-g2)/g1


def masked_percentile(regionmask, intensity):
    return np.percentile(intensity[regionmask], q=(25, 50, 75))


def masked_cv(regionmask, intensity):
    return 100*np.std(intensity[regionmask])/np.mean(intensity[regionmask])


def masked_skewness(regionmask, intensity):
    return stats.skew(intensity[regionmask])


def masked_kurtosis(regionmask, intensity):
    return stats.kurtosis(intensity[regionmask])


def masked_std(regionmask, intensity):
    return np.std(intensity[regionmask])


def gabor_kernel_bank(n_theta=4,
                      sigmas=[2],
                      lamdas = [np.pi/4],
                      gammas = [0.5],
                      kernel_size=5):
    """
    Generate a bank of Gabor filters with different orientations and parameters.
    Annotated by GPT-4 on 2023/05/21
    :param n_theta: The number of different orientations
    :type n_theta: int, optional
    :param sigmas: The standard deviation(s) of the gaussian envelope
    :type sigmas: list of float, optional
    :param lamdas: The wavelength(s) of the sinusoidal factor
    :type lamdas: list of float, optional
    :param gammas: The spatial aspect ratio(s)
    :type gammas: list of float, optional
    :param kernel_size: The size of the gabor kernel
    :type kernel_size: int, optional
    :return: A dictionary of gabor kernels
    :rtype: dict
    """
    gabor_kernels = {}
    counter=0
    for i in range(n_theta):
        theta = (i/n_theta)*np.pi
        for sigma in sigmas:
            for lamda in lamdas:
                for gamma in gammas:
                    kernel = cv2.getGaborKernel((kernel_size,kernel_size),
                                                sigma,theta,lamda,gamma,0,
                                                ktype=cv2.CV_64F)
                    gabor_kernels['Gabor{}'.format(counter)]=kernel
                    counter+=1
    return gabor_kernels

def basic_intensity_features(data):
    from scipy.stats import kurtosis,skew
    names = ['mean','median','max','min','Q1','Q3','std','CV','skewness','kurtosis']
    data_stats = [np.mean(data),
                  np.median(data),
                  np.max(data),
                  np.min(data),
                  np.percentile(data,25),
                  np.percentile(data,75),
                  np.std(data),
                  np.std(data)/np.mean(data),
                  skew(data.flatten()),
                  kurtosis(data.flatten())]
    return names, data_stats


import numpy as np
from skimage import morphology, filters, measure
from .linalg import *


"""
===================================. outlier detection .================================
"""
def zscore(data):
    """
    Calculates the Z-score (standard score, https://en.wikipedia.org/wiki/Standard_score) of input data
    :param data: list or numpy array, input n-D array
    :return: numpy array, calculated Z-score
    """
    data = np.array(data)
    return (data - data.mean()) / data.std()


def remove_outliers(df, feature='length', zcutoff=3):
    """
    :param df: pandas DataFrame
    :param feature: str, column name of the feature to remove outliers from
    :param zcutoff: float, the cutoff value for the Z-score.
                    Observations with Z-score greater than this value will be considered outliers.
    :return: pandas DataFrame, with outliers removed from the specified feature
    """
    if feature not in df:
        raise ValueError('Feature {} not found!'.format(feature))
    df_copied = df.copy().reset_index(drop=True)
    data = df[feature].values
    zscores = zscore(data)
    value_in_range = np.where(np.abs(zscores) < zcutoff)[0]
    df_filtered = df_copied.iloc[value_in_range].copy().sort_values(by=feature).reset_index(drop=True)
    return df_filtered


def remove_outliers_batch(df, features=None, zcutoffs=None):
    """
    Removes outliers from multiple features of a pandas DataFrame using the remove_outliers function.
    :param df: pandas DataFrame
    :param features: list of str, column names of the features to remove outliers from
    :param zcutoffs: list of float, the cutoff values for the Z-scores. Observations with Z-scores greater than these values will be considered outliers.
    :return: pandas DataFrame, with outliers removed from the specified features
    """

    if features is None:
        features = ['length']

    if zcutoffs is None:
        zcutoffs = [3]

    if len(features) != len(zcutoffs):
        raise ValueError("Feature and z-cutoff lists don't match")

    df_copy = df.copy()
    for f, z in zip(features, zcutoffs):
        df_copy = remove_outliers(df_copy, f, z)
    return df_copy

def find_outliers(df, feature='length', zcutoff=3):
    """
    Finds outliers in a pandas dataframe column using the Z-score method
    :param df: pandas dataframe, input dataframe
    :param feature: str, name of the column to find outliers in
    :param zcutoff: float, Z-score cutoff value
    :return: numpy array, binary array indicating whether each row is an outlier or not
    """
    if feature not in df:
        raise ValueError('Feature {} not found!'.format(feature))
    data = df[feature].values
    zscores = zscore(data)
    return (zscores > zcutoff).astype(int)


def find_outliers_batch(df, features=None, zcutoffs=None):
    """
    Finds outliers in multiple pandas dataframe columns using the Z-score method
    :param df: pandas dataframe, input dataframe
    :param features: list of str, names of the columns to find outliers in
    :param zcutoffs: list of float, Z-score cutoff values for each column
    :return: numpy array, binary array indicating whether each row is an outlier or not
    """
    if features is None:
        features = ['length']

    if zcutoffs is None:
        zcutoffs = [3]

    if len(features) != len(zcutoffs):
        raise ValueError("Feature and z-cutoff lists don't match")
    zscore_sum = np.zeros(len(df))
    for f, z in zip(features, zcutoffs):
        zscore_sum += find_outliers(df, f, z)
    return (zscore_sum>0).astype(int)


"""
===================================. CV helper .================================
"""

def gaussian_smooth(img,sigma=1,preserve_range=True):
    """
    wrapper of skimage.filters.gaussian, preserves range by default
    :param img: input image
    :param sigma: gaussian kernel sigma
    :return: gaussian smoothed image
    """
    return filters.gaussian(img,sigma=sigma,preserve_range=preserve_range)

def normalize_image(img,
                    mask=None,
                    min_perc=0.5,
                    max_perc=99.5,
                    min_v=0,
                    max_v=30000,
                    bg=0.5):
    """
    inherited from @kevinjohncutler with modifications
    perform either percentile based normalization or
    fixed range normalization or
    masked normalizatoin.

    For masked normalization:
    @ kevinjohncutler
    Normalize image by rescaling from 0 to 1 and then adjusting gamma to bring
    average background to specified value (0.5 by default).

    :params img: input two-dimensional image
    :params mask: input labels or foreground mask or anything but not none
    :params min_perc: lower bound of the percentile norm (0-100)
    :params max_perc: higher bound of the percentile norm (0-100)
    :params min_v: lower bound of the absolute value for normalization, overwrite min_perc,
    :params max_v: higher bound of the absolute value for normalization, overwrite max_perc,
    :params bg: background value in the range 0-1
    :return: gamma-normalized array with a minimum of 0 and maximum of 1
    """
    if mask is None:
        th1, th2 = np.percentile(img, min_perc), np.percentile(img, max_perc)
        if min_v is not None:
            th1 = min_v
        if max_v is not None:
            th2 = max_v
        img = (img - th1) / (th2 - th1)
        img[img > 1] = 1
        img[img < 0] = 0
        return img

    else:
        img = (img - img.min()) / (img.max() - img.min())
        try:
            img = img ** (np.log(bg) / np.log(np.mean(img[morphology.binary_erosion(mask == 0)])))
        except:
            # just in case when mask is invalid
            mask = img < filters.threshold_isodata(img)
            img = img ** (np.log(bg) / np.log(np.mean(img[morphology.binary_erosion(mask == 0)])))

    return img


def smooth_binary_mask(mask, sigma=1.5):
    """
    Applies a Gaussian filter to a binary mask for smoothing.
    :param mask: 2D numpy array representing the binary mask.
    :param sigma: Float representing the standard deviation for Gaussian kernel. Default value is 1.5.
    :return: 2D numpy array of the smoothed binary mask where mask values > 0.5 are considered True (or 1) and those <= 0.5 are considered False (or 0).
    """
    smooth_mask = filters.gaussian(mask*1, sigma=sigma,preserve_range=True)
    return smooth_mask > 0.5


def invert_normalize(data, max_percentile=99):
    """
    invert then adjust the image drange to 0-1
    :param data: image data, usually masked
    :param max_percentile: percentile maximum for normalization
    :return: inverted, normalized data
    """
    max_val = np.percentile(data, max_percentile)
    inverted = max_val - data
    inverted[inverted <= 0] = 0
    normalized = inverted / inverted.max()
    return normalized


def percentile_normalize(data, perc1=5, perc2=95, fix_lower_bound=500, fix_higher_bound=None):
    q1, q2 = np.percentile(data, perc1), np.percentile(data, perc2)
    if fix_lower_bound is not None:
        q1 = min(q1, fix_lower_bound)
    if fix_higher_bound is not None:
        q2 = max(q2, fix_higher_bound)
    float_d = data.copy().astype(float)
    transformed = (float_d - q1) / (q2 - q1)
    transformed[transformed > 1] = 1
    transformed[transformed < 0] = 0
    return transformed


def adjust_image(img, dtype=16, adjust_gamma=True, gamma=1):
    from skimage import exposure
    """
    adjust image data depth and gamma value
    :param img: input image
    :param dtype: bit depth, 8, 12 or 16
    :param adjust_gamma: whether or not correct gamma
    :param gamma: gamma value
    :return: adjusted image
    """
    n_range = (0,255)
    if isinstance(dtype, int) & (dtype > 2):
        n_range = (0, 2 ** dtype - 1)
    else:
        print("Illegal input found where an integer no less than 2 was expected.")
    outimg = exposure.rescale_intensity(img, out_range=n_range)
    if adjust_gamma:
        outimg = exposure.adjust_gamma(outimg, gamma=gamma)
    return outimg


def coord2mesh(x, y, selem,
               xmax=10000000,
               ymax=10000000):
    """

    :param x:
    :param y:
    :param selem:
    :param xmax:
    :param ymax:
    :return:
    """
    dx, dy = np.where(selem)
    dx -= int(selem.shape[0] / 2)
    dy -= int(selem.shape[1] / 2)
    x_mesh = x[:, np.newaxis] + dx[np.newaxis, :]
    x_mesh[x_mesh < 0] = 0
    x_mesh[x_mesh > xmax - 1] = xmax - 1
    y_mesh = y[:, np.newaxis] + dy[np.newaxis, :]
    y_mesh[y_mesh < 0] = 0
    y_mesh[y_mesh > ymax - 1] = ymax - 1
    return x_mesh, y_mesh


def simplify_polygon(polygon,
                     tolerance=0.95,
                      interp_distance=1,
                      min_segment_count=2):
    """
    Simplifies a given polygon using the Ramer-Douglas-Peucker algorithm and linear interpolation.

    :param polygon: A 2D numpy array of shape (n, 2) representing the polygon.
    :param tolerance: A float representing the maximum distance for a point to be considered as a candidate for reduction, default value is 0.95.
    :param interp_distance: A float representing the distance between interpolated points, default value is 1.
    :param min_segment_count: An integer representing the minimum number of interpolation points, default value is 2.
    :return: A 2D numpy array of the simplified and interpolated polygon.
    """
    from skimage.measure import approximate_polygon
    if (polygon[0]-polygon[-1]).sum() == 0:
        closed = True
    else:
        closed = False
    approximation = approximate_polygon(polygon,tolerance=tolerance)
    linear_interpolated_segments=[]
    for i in range(len(approximation)-1):
        p1 = approximation[i]
        p2 = approximation[i+1]
        dist = distance(p1,p2)
        steps = max(int(round(dist/interp_distance)),min_segment_count)
        linear_interpolated_segments.append(np.array([np.linspace(p1[0],p2[0],steps),np.linspace(p1[1],p2[1],steps)]).T[:-1])
    linear_interpolated_segments.append(np.array([approximation[-1]]))
    approximated = np.concatenate(linear_interpolated_segments)
    if int(measure_length(approximated)/interp_distance) >= 5:
        approximated = spline_approximation(approximated,
                                            n=int(measure_length(approximated)/interp_distance),
                                            closed=closed,smooth_factor=0)
    return approximated


def dilate_by_coords(coords, image_shape,
                     selem=morphology.disk(2)):
    h, w = image_shape
    if len(selem) == 1:
        return coords
    elif len(selem) > 1:
        x, y = coord2mesh(coords[:, 0],coords[:, 1],selem=selem)
        x[x < 0] = 0
        x[x > h - 1] = h - 1
        y[y < 0] = 0
        y[y > w - 1] = w - 1
        xy = np.unique(np.array([x.ravel(), y.ravel()]).T, axis=0)
        return xy


def get_neighbors(labeled_mask, label):
    """
    Retrieves the neighboring labels around a given label in a labeled image mask.

    :param labeled_mask: 2D ndarray where each unique integer represents a unique region
    :type labeled_mask: numpy.ndarray
    :param label: The label for which to find neighbors
    :type label: int
    :return: List of unique neighboring labels
    :rtype: list
    """
    _x, _y = np.where(labeled_mask == label)
    x, y = dilate_by_coords(np.array([_x, _y]).T, labeled_mask.shape).T
    return [i for i in np.unique(labeled_mask[x, y]) if i not in [0, label]]


"""
===================================. miscellaneous .================================
"""
def dict2list(dict_f):
    """
    This function takes a dictionary as input and returns a list of lists,
    where each inner list contains a key-value pair from the dictionary.

    :param dict_f: input dictionary
    :return: output list
    """
    return [[k, v] for k, v in dict_f.items()]


def identical_shapes(image_list):
    """
    Check if all images in the list have identical shapes.
    :param image_list: A list of numpy arrays representing images.
    :return: True if all images have identical shapes, False otherwise.
    """
    x, y = 0, 0
    shape_matched = True
    for i, img in enumerate(image_list):
        if i == 0:
            x, y = img.shape
        else:
            if x != img.shape[0] or y != img.shape[1]:
                shape_matched = False
    return shape_matched


def value_constrain(value, min_value, max_value):
    """
    Keeps the value between a user defined range.

    :param value: The value to be constrained.
    :param min_value: The minimum value.
    :param max_value: The maximum value.
    :return: The constrained value.
    """
    return max(min_value, min(value, max_value))


def in_range(val, min_val, max_val):
    """
    check whether a value is within a user defined range
    :param val:
    :param min_val:
    :param max_val:
    :return:
    """
    if (val < min_val) or (val > max_val):
        return False
    else:
        return True


def norm_ratio(measure_of_two):
    """
    Calculate the normalized ratio of two values.
    Basically it does: min(measure_of_two) / max(measure_of_two)

    :param measure_of_two: The two values to be normalized.
    :return: The normalized ratio of the two values.
    """
    sorted_v = np.sort(measure_of_two)
    return sorted_v[0] / sorted_v[1]


def kwarg_or_default(key, kwarg, default):
    """
    Get the value of a keyword argument or return a default value.

    :param key: str, the key of the keyword argument.
    :param kwarg: dict, the dictionary of keyword arguments.
    :param default: The default value to return if the key is not in the dictionary.
    :return: The value of the keyword argument or the default value.
    """
    if key in kwarg:
        return kwarg[key]
    else:
        return default



def reorient_data(data, skew='left'):
    skewed_direction = 'right'
    interpolated = np.interp(np.linspace(0, 1, 100), np.linspace(0, 1, len(data)), data)
    if interpolated[:50].mean() > interpolated.mean():
        skewed_direction = 'left'
    if skewed_direction == skew:
        return False
    else:
        return True


def current_datetime():
    from datetime import datetime
    # datetime object containing current date and time
    now = datetime.now()
    # dd/mm/YY H:M:S
    dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
    return dt_string


def rank2group(n, n_group=50):
    """
    Divides a range of numbers into groups. The numbers range from 0 to n-1.
    The groups are created in a way to divide the range into equal parts as much as possible.

    :param n: The total number of elements
    :type n: int, optional
    :param n_group: The total number of groups
    :type n_group: int, optional
    :return: A list of numpy arrays where each array represents a group of indices
    :rtype: list
    """

    ext = int(0.5 * (n % n_group))
    step = int(n / n_group)
    indices = []
    for i in range(n_group):
        group = np.arange(i * step, i * step + step) + ext
        group = group[group < n - 1]
        indices.append(group)
    return indices


def min_max(data):
    """
    Min-max normalization of n-D data
    :param data: list or numpy array.
    :return: min-max normalized numpy array.
    """
    data = np.array(data)
    return (data - np.min(data)) / (np.max(data) - np.min(data))


def correlation_map(x, y):
    """
    Correlate each n with each m.
    @abcd
    @https://stackoverflow.com/questions/30143417/computing-the-correlation-coefficient-between-two-multi-dimensional-arrays/30145770#30145770
    :param x: numpy array, Shape N X T.
    :param y: numpy array, Shape M X T.
    :return: numpy array, N X M array in which each element is a correlation coefficient.
    """
    mu_x = x.mean(1)
    mu_y = y.mean(1)
    n = x.shape[1]
    if n != y.shape[1]:
        raise ValueError('x and y must have the same number of timepoints.')
    s_x = x.std(1, ddof=n-1)
    s_y = y.std(1, ddof=n-1)
    cov = np.dot(x,y.T) - n * np.dot(mu_x[:, np.newaxis],mu_y[np.newaxis, :])
    return cov/np.dot(s_x[:, np.newaxis], s_y[np.newaxis, :])


def dist_prob(x, d=0.1, t=1, a=0.1):
    """
    Gaussian distance probability distribution

    :param x: float, input value
    :param d: float, diffusion coefficient (default=0.1)
    :param t: float, time (default=1)
    :param a: float, constant (default=0.1)
    :return: float, probability density
    """
    return (1 / np.sqrt(4 * np.pi * t * d)) * np.exp(-(a * x) ** 2 / (4 * d * t))


def difference_matrix(v1, v2, ref=0):
    """
    Computes the normalized difference matrix between two vectors.
    The normalized difference is computed as the absolute difference divided by the mean.
    This results in a similarity score where a score of 1 means the vectors are identical,
    and a score of 0 means they are completely dissimilar.

    :param v1: First input vector
    :type v1: numpy.ndarray
    :param v2: Second input vector
    :type v2: numpy.ndarray
    :param ref: Reference value to initialize the difference matrix, default is 0
    :type ref: int or float, optional
    :return: Normalized difference matrix
    :rtype: numpy.ndarray
    """
    diff = np.abs(v1[:, np.newaxis, :] - v2[np.newaxis, :, :])
    mean = (v1[:, np.newaxis, :] + v2[np.newaxis, :, :]) / 2
    norm_similarity = 1 - np.mean(diff / mean, axis=-1)
    norm_similarity[norm_similarity < 0] = 0
    return norm_similarity




from .contour import *
import warnings


def midline_approximation(skeleton,
                          smoothed_contour,
                          move_pole1=False,
                          move_pole2=False,
                          anchor_length=3,
                          tolerance=0.1, max_iteration=10):
    """
    Approximation of a smooth midline with skeleton and a smoothed contour.

    :param skeleton: A numpy array of shape (n, 2) representing the skeleton of a 2D object.
    :param smoothed_contour: A numpy array of shape (m, 2) representing the smoothed contour of a 2D object.
    :param tolerance: A float representing the maximum distance between the midline and the skeleton at which convergence is achieved.
    :param max_iteration: An integer representing the maximum number of iterations to perform.
    :return: A tuple containing a numpy array of shape (p, 2) representing the midline approximation and a boolean indicating whether convergence was achieved.
    """
    midline = skeleton.copy()
    pole1_anchor = []
    pole2_anchor = []

    if not move_pole1:
        pole1_anchor = midline[:anchor_length]
        midline = midline[anchor_length:]
    if not move_pole2:
        pole2_anchor = midline[-anchor_length:]
        midline = midline[:-anchor_length]

    n = 0
    converged = False
    while n < max_iteration:
        updated_midline = skeleton_contour_intersect_points(midline, smoothed_contour)
        dxy = updated_midline - midline
        midline = spline_approximation(updated_midline,
                                               n=len(updated_midline),
                                               smooth_factor=1,
                                       closed=False)
        if np.abs(dxy).mean() <= tolerance:
            converged = True
            break
        n += 1
    stiched_midline = []
    if not move_pole1:
        stiched_midline.append(pole1_anchor)
    stiched_midline.append(updated_midline)
    if not move_pole2:
        stiched_midline.append(pole2_anchor)
    midline = np.vstack(stiched_midline)
    return midline.astype(float), converged


def skeleton_analysis(mask,
                      pruning=False,
                      min_branch_length=5,
                      max_iterations=30,
                      method='zhang'):
    """
    find all skeleton branch(es) from a binary mask and sort the pixel coordinates accordingly
    :param mask:
    :param pruning:
    :param min_branch_length:
    :param max_iterations:
    :param method:
    :return:
    """
    warnings.filterwarnings("ignore")
    skeleton = morphology.skeletonize(mask,method=method)*2
    endpoints, branch_points, skeleton = locate_nodes(skeleton)
    anchor_points = endpoints+branch_points
    skeleton_branches = []
    n = 0
    if (len(endpoints) >= 2): #and (len(branch_points) <= 10):
        while True:
            if len(anchor_points) == 0:
                break
            if n>=max_iterations:
                break
            is_real_pole = [0,0]
            startpoint = anchor_points[0]
            if startpoint in endpoints:
                is_real_pole[0] = 1
            xcoords,ycoords = neighbor_search(skeleton,startpoint)
            anchor_points.remove(startpoint)
            if len(xcoords)>=1:
                lx,ly = xcoords[-1],ycoords[-1]
                node_count, node_coords = cubecount_with_return(skeleton, lx, ly, 3)
                if (node_count == 0) and ([lx,ly] in endpoints):
                    is_real_pole[1] = 1
                    anchor_points.remove([lx, ly])
                elif (node_count == 1):
                    lx, ly = node_coords[0]
                    if [lx, ly] in anchor_points:
                        anchor_points.remove([lx, ly])
                    xcoords.append(lx)
                    ycoords.append(ly)
                else:
                    for xy in node_coords:
                        if distance([lx,ly],xy) == 1:
                            lx,ly = xy
                            break
                    if [lx, ly] in anchor_points:
                        anchor_points.remove([lx, ly])
                    xcoords.append(lx)
                    ycoords.append(ly)
                skeleton_branches.append([is_real_pole, xcoords, ycoords])
            n+=1
        branch_copy = skeleton_branches.copy()
        if pruning:
            for branch in branch_copy:
                if len(branch[1]) <= min_branch_length:
                    skeleton_branches.remove(branch)
            if len(skeleton_branches)==1:
                skeleton_branches[0][0]=[1,1]
        if n < max_iterations:
            return skeleton_branches, skeleton
        else:

            return [], skeleton
    else:
        return [], skeleton

def locate_nodes(skeleton):
    """
    traverse the skeleton points and locate branch and terminal pixels.
    :param skeleton: binary skeleton generated by calling skimage.morphology.skeletonize
    :return: a list of three short lists of coordinates:
             @endpoints: outreaching pole coordinates, there should be one for each branch
             @branch_points: branching point coordinates, there should be one for each branch
             @skeleton: labeled skeleton
    """
    warnings.filterwarnings("ignore")
    endpoints = []
    branch_points = []
    skeleton_path = np.where(skeleton > 0)
    skeleton_length = len(skeleton_path[0])
    if skeleton_length > 5:
        for i in range(skeleton_length):
            x = skeleton_path[0][i]
            y = skeleton_path[1][i]
            if cube_nonzero(skeleton, x, y) == 1:
                endpoints.append([x, y])
            if cube_nonzero(skeleton, x, y) == 2:
                _n, _neighbors = cube_nonzero_with_return(skeleton, x, y)
                _x0, _y0 = _neighbors[0]
                _x1, _y1 = _neighbors[1]
                dist = abs(_x0 - _x1) + abs(_y0 - _y1)
                if dist == 1:
                    skeleton[x, y] = 0
                    if skeleton[_x0, _y0] == 3:
                        skeleton[_x0, _y0] = 2
                    if skeleton[_x1, _y1] == 3:
                        skeleton[_x1, _y1] = 2
                    if [_x0, _y0] in branch_points:
                        branch_points.remove([_x0, _y0])
                    if [_x1, _y1] in branch_points:
                        branch_points.remove([_x1, _y1])
            if cube_nonzero(skeleton, x, y) > 2:
                branch_points.append([x, y])
                skeleton[x, y] = 3
        return endpoints, branch_points, skeleton
    else:
        return endpoints, branch_points, skeleton


def merge_branch_points(skeleton_coords,
                        max_dist=3):
    centroid_groups_id = []
    centroid_groups_x = []
    centroid_groups_y = []
    centroid_groups = []

    for p,x,y in skeleton_coords:
        if p[0]==0:
            centroid_groups_x.append(x[0])
            centroid_groups_y.append(y[0])
        if p[1]==0:
            centroid_groups_x.append(x[-1])
            centroid_groups_y.append(y[-1])
    centroid_groups_x=np.array(centroid_groups_x)
    centroid_groups_y=np.array(centroid_groups_y)
    new_coords = []
    for p,x,y in skeleton_coords:
        new_x = x.copy()
        new_y = y.copy()
        if p[0]==0:
            new_x[0]=centroid_groups_x[np.abs(centroid_groups_x-x[0])<max_dist].mean()
            new_y[0]=centroid_groups_y[np.abs(centroid_groups_y-y[0])<max_dist].mean()
        if p[1]==0:
            new_x[-1]=centroid_groups_x[np.abs(centroid_groups_x-x[-1])<max_dist].mean()
            new_y[-1]=centroid_groups_y[np.abs(centroid_groups_y-y[-1])<max_dist].mean()
        new_coords.append([p,new_x,new_y])
    return new_coords

@jit(nopython=True, cache=True)
def neighbor_search(input_map, endpoint, max_iterations=500):
    output_x = []
    output_y = []
    end_reached = False
    x, y = endpoint[0], endpoint[1]
    counter = 0
    while not end_reached:
        if counter >= max_iterations:
            break
        n_neighbor, neighbor_list = cubecount_with_return(input_map, x, y, 2)
        if n_neighbor == 0:
            input_map[x, y] = 1
            output_x.append(x)
            output_y.append(y)
            end_reached = True
        elif n_neighbor == 1:
            input_map[x, y] = 1
            output_x.append(x)
            output_y.append(y)
            x, y = neighbor_list[-1]
        elif n_neighbor >= 2:
            input_map[x, y] = 1
            output_x.append(x)
            output_y.append(y)
            end_reached = True
        counter += 1
    return output_x, output_y


@jit(nopython=True, cache=True)
def cubesum(mask, px, py, val):
    n = 0
    for i in [0, 1, 2, 3, 5, 6, 7, 8]:
        dx = int(i / 3)
        dy = i - dx * 3
        x1, y1 = px + dx - 1, py + dy - 1
        if mask[x1, y1] == val:
            n += 1
    return n


@jit(nopython=True, cache=True)
def cube_nonzero(mask, px, py):
    n = 0
    for i in [0, 1, 2, 3, 5, 6, 7, 8]:
        dx = int(i / 3)
        dy = i - dx * 3
        x1, y1 = px + dx - 1, py + dy - 1
        if mask[x1, y1] > 0:
            n += 1
    return n


@jit(nopython=True, cache=True)
def cube_nonzero_with_return(mask, px, py):
    retlist = []
    n = 0
    for i in [0, 1, 2, 3, 5, 6, 7, 8]:
        dx = int(i / 3)
        dy = i - dx * 3
        x1, y1 = px + dx - 1, py + dy - 1
        if mask[x1, y1] > 0:
            retlist.append([x1, y1])
            n += 1
    return n, retlist


@jit(nopython=True, cache=True)
def cubecount_with_return(mask, px, py, val):
    retlist = []
    n = 0
    for i in [0, 1, 2, 3, 5, 6, 7, 8]:
        dx = int(i / 3)
        dy = i - dx * 3
        x1, y1 = px + dx - 1, py + dy - 1
        if mask[x1, y1] == val:
            retlist.append([x1, y1])
            n += 1
    return n, retlist


def find_poles(smoothed_skeleton,
               smoothed_contour,
               find_pole1=True,
               find_pole2=True,
               gap=4, ):
    """

    :param smoothed_skeleton:
    :param smoothed_contour:
    :param find_pole1:
    :param find_pole2:
    :param gap:
    :return:
    """
    # find endpoints and their nearest neighbors on a midline
    length = len(smoothed_skeleton)
    extended_pole1 = [smoothed_skeleton[0]]
    extended_pole2 = [smoothed_skeleton[-1]]
    i = 0
    j = 0
    if find_pole1:
        for i in range(5):
            p1 = smoothed_skeleton[i]
            p2 = smoothed_skeleton[i + gap]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole1 = line_contour_intersection(p1, p2, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_1 = p1 - p2
            p1_tile = np.tile(p1, (len(intersection_points_pole1), 1))
            p1dot = (intersection_points_pole1 - p1_tile) * dxy_1
            index_1 = np.where((p1dot[:, 0] > 0) & (p1dot[:, 1] > 0))[0]
            if len(index_1) > 0:
                extended_pole1 = intersection_points_pole1[index_1]
                break
    else:
        i = 1

    if find_pole2:
        for j in range(5):
            p3 = smoothed_skeleton[-1 - j]
            p4 = smoothed_skeleton[-1 - gap - j]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole2 = line_contour_intersection(p3, p4, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_2 = p3 - p4
            p3_tile = np.tile(p3, (len(intersection_points_pole2), 1))
            p3dot = (intersection_points_pole2 - p3_tile) * dxy_2
            index_2 = np.where((p3dot[:, 0] > 0) & (p3dot[:, 1] > 0))[0]
            if len(index_2) > 0:
                extended_pole2 = intersection_points_pole2[index_2]
                break
    else:
        j = 1
    trimmed_midline = smoothed_skeleton[i:length - j]
    return extended_pole1, extended_pole2, trimmed_midline


def find_poles_simp(smoothed_skeleton,
                    smoothed_contour,
                    find_pole1=True,
                    find_pole2=True,
                    gap=3):
    """
    find endpoints and their nearest neighbors on a midline
    :param smoothed_skeleton:
    :param smoothed_contour:
    :param find_pole1:
    :param find_pole2:
    :param gap:
    :return:
    """

    length = len(smoothed_skeleton)
    extended_pole1 = [smoothed_skeleton[0]]
    extended_pole2 = [smoothed_skeleton[-1]]
    i = 0
    j = 0
    if find_pole1:
        for i in range(5):
            p1 = smoothed_skeleton[i]
            p2 = smoothed_skeleton[i + gap]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole1 = line_contour_intersection(p1, p2, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_1 = p1 - p2
            p1_tile = np.tile(p1, (len(intersection_points_pole1), 1))
            p1dot = (intersection_points_pole1 - p1_tile) * dxy_1
            index_1 = np.where((p1dot[:, 0] > 0) & (p1dot[:, 1] > 0))[0]
            if len(index_1) > 0:
                extended_pole1 = intersection_points_pole1[index_1]
                break
    else:
        i = 1

    if find_pole2:
        for j in range(5):
            p3 = smoothed_skeleton[-1 - j]
            p4 = smoothed_skeleton[-1 - gap - j]
            # find the two intersection points between
            # the vectorized contour and line through pole1
            intersection_points_pole2 = line_contour_intersection(p3, p4, smoothed_contour)
            # find the interesection point with the same direction as the outward pole vector
            dxy_2 = p3 - p4
            p3_tile = np.tile(p3, (len(intersection_points_pole2), 1))
            p3dot = (intersection_points_pole2 - p3_tile) * dxy_2
            index_2 = np.where((p3dot[:, 0] > 0) & (p3dot[:, 1] > 0))[0]
            if len(index_2) > 0:
                extended_pole2 = intersection_points_pole2[index_2]
                break
    else:
        j = 1
    trimmed_midline = smoothed_skeleton[i:length - j]
    return extended_pole1, extended_pole2, trimmed_midline


def extend_skeleton(smoothed_skeleton,
                    smoothed_contour,
                    find_pole1=True,
                    find_pole2=True,
                    subdivide_pole_segment=3,
                    upsampling_factor=1):
    # initiate approximated tip points
    new_pole1, new_pole2, smoothed_skeleton = find_poles(smoothed_skeleton,
                                                         smoothed_contour,
                                                         find_pole1=find_pole1,
                                                         find_pole2=find_pole2)

    extended_skeleton = []
    if find_pole1:
        segment1 = (np.array([np.linspace(new_pole1[0][0], smoothed_skeleton[0, 0], subdivide_pole_segment),
                              np.linspace(new_pole1[0][1], smoothed_skeleton[0, 1], subdivide_pole_segment)]).T)[:-1]
    else:
        segment1 = new_pole1
    if find_pole2:
        segment2 = (np.array([np.linspace(smoothed_skeleton[-1, 0], new_pole2[0][0], subdivide_pole_segment),
                              np.linspace(smoothed_skeleton[-1, 1], new_pole2[0][1], subdivide_pole_segment)]).T)[1:]
    else:
        segment2 = new_pole2

    extended_skeleton = np.concatenate([segment1,
                                        smoothed_skeleton,
                                        segment2])
    l = int(upsampling_factor*round(measure_length(smoothed_skeleton)))
    return spline_approximation(extended_skeleton,
                                n=l,
                                smooth_factor=1, closed=False)


def fast_midline(mask, outline,
                 centroid=None,
                 orientation=None,
                 pixel_microns=0.065,
                 pole_length=0.3,
                 upsampling_factor=2,
                 tolerance = 0.1,
                 max_iteration = 10):
    """

    """

    from skimage.measure import regionprops
    # get centroid coordinates
    if centroid is None:
        mask_x, mask_y = np.where(mask == 1)
        x0, y0 = np.round(np.array([mask_x.mean(), mask_y.mean()])).astype(int)
    else:
        x0, y0 = np.round(centroid).astype(int)
    # get particle orientation
    if orientation is None:
        orientation = regionprops(mask * 1)[0].orientation

    # consider both poles as floppy ends
    pole1, pole2 = 1, 1

    # get the line segment that traverses the centroid and aligns with the orientation of the particle
    dx, dy = np.cos(orientation), np.sin(orientation)
    intersect = line_contour_intersection([x0, y0],
                                          [x0 + dx, y0 + dy],
                                              outline)
    # estimate interpolation size
    l = int(round(measure_length(intersect) * upsampling_factor))
    pole_length_pix = int(np.round(upsampling_factor * pole_length / pixel_microns)) + 1

    # linear interpolation to generate the initial midline
    init_midline = np.array([np.linspace(intersect[0, 0], intersect[1, 0], l),
                             np.linspace(intersect[0, 1], intersect[1, 1], l)]).T

    # use only the middle part of the interpolated line to fit the cell midline
    mid_segment = init_midline[pole_length_pix * pole1:-pole_length_pix * pole2]
    smooth_midline, _converged = midline_approximation(mid_segment,
                                                        outline,
                                                        move_pole1=pole1,
                                                        move_pole2=pole2,
                                                        tolerance=tolerance,
                                                        max_iteration=max_iteration,
                                                        anchor_length=3)

    # if converted, extend the centerline to
    if _converged:
        midline = extend_skeleton(smooth_midline,
                                   outline,
                                   find_pole1=pole1,
                                   find_pole2=pole2)
    else:
        midline = init_midline
    midline = spline_approximation(midline, n=max(3, int(round(measure_length(midline)))),
                                   smooth_factor=1, closed=False)
    return [midline]


def extract_midline(mask,
                    outline,
                    pixel_microns=0.065,
                    min_branch_length=0.2,
                    method='zhang'):
    """
    """
    from skimage import morphology
    min_branch_length_pix = max(3,int(round(min_branch_length/pixel_microns)))
    skeleton_coords, skeleton = skeleton_analysis(morphology.binary_opening(mask),
                                                   pruning=True,
                                                   method=method,
                                                   min_branch_length=min_branch_length_pix)
    skeleton_coords = merge_branch_points(skeleton_coords,max_dist=3)
    midline_coords = []
    for i in range(len(skeleton_coords)):
        pole1, pole2 = skeleton_coords[i][0]
        xcoords = skeleton_coords[i][1]
        ycoords = skeleton_coords[i][2]
        skel_coords = np.array([xcoords, ycoords]).T
        if len(skel_coords)<15:
            interp_distance = 0.5
        elif len(skel_coords)<60:
            interp_distance = 1
        elif len(skel_coords)<90:
            interp_distance = 2
        else:
            interp_distance = 3
        simp = simplify_polygon(skel_coords,
                                         tolerance=0.9,
                                         interp_distance=interp_distance)
        extended = extend_skeleton(simp,
                                       outline,
                                       find_pole1=pole1,
                                       find_pole2=pole2,
                                       subdivide_pole_segment=5,
                                       upsampling_factor=1/interp_distance)
        smooth_skel, _converged = midline_approximation(extended,
                                                        outline,
                                                        max_iteration=5,
                                                        tolerance=0.15,
                                                        move_pole1 = False,
                                                        move_pole2 = False,
                                                        anchor_length=5)
        if _converged:
            smooth_skel=simplify_polygon(smooth_skel,tolerance=0.9,
                                          interp_distance=interp_distance)
        else:
            smooth_skel=simplify_polygon(extended,tolerance=0.9,
                                          interp_distance=interp_distance)
        smooth_skel=spline_approximation(smooth_skel,
                                          n=int(round(measure_length(skel_coords))),
                                          smooth_factor=1,closed=False)
        midline_coords.append([[pole1,pole2],smooth_skel[:,0],smooth_skel[:,1]])
    midline_coords = merge_branch_points(midline_coords,max_dist=3)
    return [np.array([x[1],x[2]]).T for x in midline_coords]



def skeleton_contour_intersect_distance(skeleton, contour):
    """
    Compute the distance between the skeleton and the contour at their direct intersection points.
    Core function for midline approximation.
    noted 2023.04.12: renamed from "direct_intersect_distance" to "skeleton_contour_intersect_distance".

    :param skeleton: A numpy array of shape (n, 2) representing the skeleton of a 2D object.
    :param contour: A numpy array of shape (m, 2) representing the contour of a 2D object.
    :return: A numpy array of shape (p,) representing the distances between the skeleton and the contour at their direct intersection points.
    """
    v1, v2 = contour[:-1], contour[1:]
    skel_x, skel_y = skeleton[1:-1].T
    intersect_x, intersect_y = line_contour_intersect_matrix(skeleton[1:-1], contour)
    dx_v1 = intersect_x - v1.T[0][:, np.newaxis]
    dx_v2 = intersect_x - v2.T[0][:, np.newaxis]
    dy_v1 = intersect_y - v1.T[1][:, np.newaxis]
    dy_v2 = intersect_y - v2.T[1][:, np.newaxis]
    dx = dx_v1 * dx_v2
    dy = dy_v1 * dy_v2
    dist_x = skel_x[np.newaxis, :] - intersect_x
    dist_y = skel_y[np.newaxis, :] - intersect_y

    non_boundry_points = np.where(np.logical_and(dy > 0, dx > 0))
    dist_matrix = np.sqrt(dist_x ** 2 + dist_y ** 2)
    dist_matrix[non_boundry_points] = np.inf
    nearest_id_x = np.argsort(dist_matrix, axis=0)[:2]
    nearest_id_y = np.linspace(0, dist_matrix.shape[1] - 1, dist_matrix.shape[1]).astype(int)
    dists = dist_matrix[nearest_id_x[0], nearest_id_y] + dist_matrix[nearest_id_x[1], nearest_id_y]
    return np.concatenate([[0], dists, [0]])


def skeleton_contour_intersect_points(skeleton, contour):
    """
    Compute the midpoint between the skeleton and the contour at their direct intersection points.

    :param skeleton: A numpy array of shape (n, 2) representing the skeleton of a 2D object.
    :param contour: A numpy array of shape (m, 2) representing the contour of a 2D object.
    :return: A numpy array of shape (p, 2) representing the midpoints between the skeleton and the contour at their direct intersection points.
    """
    v1, v2 = contour[:-1], contour[1:]
    skel_x, skel_y = skeleton.T
    intersect_x, intersect_y = line_contour_intersect_matrix(skeleton, contour)
    dx_v1 = intersect_x - v1.T[0][:, np.newaxis]
    dx_v2 = intersect_x - v2.T[0][:, np.newaxis]
    dy_v1 = intersect_y - v1.T[1][:, np.newaxis]
    dy_v2 = intersect_y - v2.T[1][:, np.newaxis]
    dx = dx_v1 * dx_v2
    dy = dy_v1 * dy_v2
    dist_x = skel_x[np.newaxis, :] - intersect_x
    dist_y = skel_y[np.newaxis, :] - intersect_y

    non_boundry_points = np.where(np.logical_and(dy > 0, dx > 0))
    dist_matrix = np.sqrt(dist_x ** 2 + dist_y ** 2)
    dist_matrix[non_boundry_points] = np.inf
    nearest_id_x = np.argsort(dist_matrix, axis=0)[:2]
    nearest_id_y = np.linspace(0, dist_matrix.shape[1] - 1, dist_matrix.shape[1]).astype(int)

    pos1_list = np.array([intersect_x[nearest_id_x[0], nearest_id_y], intersect_y[nearest_id_x[0], nearest_id_y]]).T
    pos2_list = np.array([intersect_x[nearest_id_x[1], nearest_id_y], intersect_y[nearest_id_x[1], nearest_id_y]]).T
    midpoints = (pos1_list + pos2_list) / 2
    d_midpoints = np.abs(midpoints - skeleton)
    outliers = np.where(d_midpoints > 1)
    midpoints[outliers] = skeleton[outliers]
    return midpoints


def measure_along_strip(line, img, width = 1, subpixel = 0.5 ,mode='mean'):
    """
    Measure the intensity along a strip perpendicular to a line in an image.
    :param line: A numpy array of shape (n, 2) representing the line along which to measure intensity.
    :param img: A numpy array of shape (m, p) representing the image in which to measure intensity.
    :param width: A float representing the width of the strip perpendicular to the line.
    :param subpixel: A float representing the subpixel resolution of the strip.
    :param mode: A string representing the method for aggregating intensity values. Options are 'mean', 'max', 'min', and 'median'.
    :return: A numpy array of shape (p,) representing the intensity values along the strip.
    """
    unit_dxy = unit_perpendicular_vector(line, closed=False)
    width_normalized_dxy = unit_dxy * subpixel
    copied_img = img.copy()
    data = bilinear_interpolate_numpy(copied_img, line.T[0], line.T[1])
    for i in range(1, 1+ int(width * 0.5 / subpixel)):
        dxy = width_normalized_dxy * i
        v1 = line + dxy
        v2 = line - dxy
        p1 = bilinear_interpolate_numpy(copied_img, v1.T[0], v1.T[1])
        p2 = bilinear_interpolate_numpy(copied_img, v2.T[0], v2.T[1])
        data = np.vstack([p1, data, p2])

    if mode in ['MAX', 'max', 'max_value']:
        return np.max(data, axis=0)
    elif mode in ['MIN', 'min', 'min_value']:
        return np.min(data, axis=0)
    elif mode in ['MEDIAN', 'Med', 'median', 'med', 'MED']:
        return np.median(data, axis=0)
    else:
        return np.average(data, axis=0)


def orthogonal_mesh(midline, contour, width_list,
                    length=None,
                    median_width=None, scale=2):
    """

    """
    if type(length) not in [float, int]:
        length = measure_length(midline)
    if type(median_width) not in [float, int]:
        median_width = np.median(width_list)

    n_length = int(len(midline) * scale)
    n_width = int(len(midline) * median_width * scale / length)

    edge_points = []
    subpixel_midline = spline_approximation(midline, n=n_length,
                                            smooth_factor=1, closed=False)
    unit_perp = unit_perpendicular_vector(subpixel_midline)

    for i in range(1, len(subpixel_midline) - 1):
        # make core mesh
        p = line_contour_intersection(subpixel_midline[i],
                                      subpixel_midline[i] + unit_perp[i], contour)
        flip = point_line_orientation(subpixel_midline[i], subpixel_midline[i + 1], p[0])
        if flip == -1:
            p = np.flip(p, axis=0)
        edge_points.append(p.flatten())

    edge_points = np.array(edge_points)

    x, y = edge_points[:, 0], edge_points[:, 1]
    dxy = (edge_points[:, 2:] - edge_points[:, :2]) / n_width
    mat = np.tile(np.arange(n_width + 1), (len(edge_points), 1))
    mat_x = x[:, np.newaxis] + mat * dxy[:, 0][:, np.newaxis]
    mat_y = y[:, np.newaxis] + mat * dxy[:, 1][:, np.newaxis]

    return np.array([mat_x, mat_y])




from ._helper import *
from ..utils import *
import pandas as pd

__all__ = ['BinarizeLegend','Patchifier','MakeMask']

def mask_custom(img, th1=0, th2='auto',
                return_threshold=False, config=None, ):
    if th2 == 'auto':
        th2 = filters.threshold_isodata(img)
    binary = ((img < th2) & (img > th1)) * 1
    if not return_threshold:
        return binary.astype(int)
    else:
        return binary.astype(int), (th1, th2)


def mask_sauvola(img,
                 dark_foreground=True,
                 return_threshold=False,
                 config=None,
                 **kwargs):
    th = filters.threshold_sauvola(img, **kwargs)
    if dark_foreground:
        binary = (img < th) * 1
    else:
        binary = (img > th) * 1
    if return_threshold:
        return binary.astype(int), th
    else:
        return binary.astype(int)


def mask_isodata(img,
                 dark_foreground=True,
                 return_threshold=False,
                 config=None,
                 **kwargs):
    th = filters.threshold_isodata(img, **kwargs)
    if dark_foreground:
        binary = (img < th) * 1
    else:
        binary = (img > th) * 1
    if return_threshold:
        return binary.astype(int), th
    else:
        return binary.astype(int)


def mask_local(img,
               dark_foreground=True,
               return_threshold=False,
               block_size=15,
               config=None,
               **kwargs):
    th = filters.threshold_local(img, block_size=block_size, **kwargs)
    if dark_foreground:
        binary = (img < th) * 1
    else:
        binary = (img > th) * 1
    if return_threshold:
        return binary.astype(int), th
    else:
        return binary.astype(int)


def mask_legacy(img,
                block_sizes=(15, 81),
                method='gaussian',
                dilation_selem=morphology.disk(1),
                opening_selem=morphology.disk(1),
                config=None):
    """
    binarization method used in OMEGA (MOMIA1), deprecated.
    """
    local_mask1 = mask_local(img, block_sizes[0], method=method)
    local_mask2 = mask_local(img, block_sizes[1], method=method)
    glob_mask = mask_isodata(img)
    glob_mask = morphology.binary_dilation(glob_mask, dilation_selem)
    binary = ((local_mask1 + local_mask2) * glob_mask) > 0
    binary = morphology.binary_opening(binary, opening_selem)
    return binary.astype(int)


def mask_composite(img, config=None):
    # reset configurations when provided
    if config is not None:
        window_size = int(config['segment']['sauvola_window_size'])
        k = float(config['segment']['sauvola_k'])
        min_particle_size = int(config['segment']['min_size'])
        opening = bool(int(config['segment']['binary_opening']))
        overlap_threshold = float(config['segment']['overlap_threshold'])
        dilation = bool(int(config['segment']['binary_dilation']))
    else:
        window_size = 9  # sauvola local threshold window size
        k = 0.05  # sauvola local threshold regulation
        min_particle_size = 10  # minimal particle size included
        opening = False  # opening to delink neighboring particles
        overlap_threshold = 0.5  # threshold of overlap between masks rendered by local and global thresholding
        dilation = True

    # local threshold
    mask1 = mask_sauvola(img, window_size=window_size, k=k)
    # isodata threshold
    mask2 = mask_isodata(img)
    # composite mask
    init_mask = morphology.remove_small_objects((mask1 + mask2) > 0, min_size=min_particle_size)
    if opening:
        init_mask = morphology.binary_opening(init_mask)

    # label composite mask
    labels = measure.label(init_mask * 1)
    filtered_labels = labels.copy()

    # measure particles, define foreground/background
    region_props = pd.DataFrame(
        measure.regionprops_table(labels, intensity_image=mask2, properties=['coords', 'mean_intensity']))
    foreground = (region_props['mean_intensity'] >= overlap_threshold) * 1
    region_props['foreground'] = foreground

    # remove background particles
    x, y = np.vstack(region_props[foreground == 0]['coords'].values).T
    filtered_labels[x, y] = 0
    binary = (filtered_labels > 0) * 1
    if dilation:
        binary = morphology.binary_dilation(binary)
    return binary.astype(int)


def _make_mask(img, method=3, **kwargs):
    if method == 0:
        return mask_custom(img, **kwargs)
    if method == 1:
        return mask_isodata(img, **kwargs)
    if method == 2:
        return mask_sauvola(img, **kwargs)
    if method == 3:
        return mask_local(img, **kwargs)
    if method == 4:
        return mask_legacy(img, **kwargs)
    if method == 5:
        return mask_composite(img, **kwargs)


class Patchifier:
    """
    A simple way to convert 2D images to patches and stitch them back into one
    Currently it only works with images with shapes like (height, width) or (height, width, channel), it doesn't work
    on image series such as (frame, height, width, channel)
    The smoothing function for overlap edges is simply the mean values of the overlapping pixels. Future updates may
    consider implementing 2D spline interpolation based smoothing method, such as:
    https://github.com/bnsreenu/python_for_microscopists/blob/master/229_smooth_predictions_by_blending_patches/smooth_tiled_predictions.py
    """

    def __init__(self, img_shape=(512, 512), patch_size=128, pad=32):

        """
        :param img_shape: the original shape of the large input image
        :param patch_size: the height and width of the square-shaped patch
        :param pad: half-width of the overlapping region of neighboring patches
        """

        self._shape = img_shape
        self.size = patch_size
        self.pad = pad
        self.shape = (max(self._shape[0], self.size),
                      max(self._shape[1], self.size))
        self.pad_h = max(0, self.size - self._shape[0])
        self.pad_w = max(0, self.size - self._shape[1])
        self.ref_coords = self.generate_patch_coords()

    def generate_patch_coords(self):
        """
        funtion to generate patch coords
        :return:
        """
        h, w = self.shape
        xs = list(np.arange(0, h - self.size, self.size - 2 * self.pad)) + [h - self.size]
        if len(xs) > 1:
            if xs[-1] == xs[-2]:
                xs = xs[:-1]
        ys = list(np.arange(0, w - self.size, self.size - 2 * self.pad)) + [w - self.size]
        if len(ys) > 1:
            if ys[-1] == ys[-2]:
                ys = ys[:-1]
        ref_coords = np.array([[x, y, np.random.randint(2)] for x in xs for y in ys])
        return ref_coords

    def pachify(self, img, random_rotate=False):
        """
        convert img to patches
        :param img: MOMIA2 image
        :param random_rotate: if randomly rotate clips, this shouldn't be used for prediction but can be helpful for training
        :return: clipped patches
        """
        if self.shape != img.shape[:2]:
            self.__init__(img.shape[:2])
        pad_config = np.zeros((len(img.shape), 2))
        pad_config[0][1] = self.pad_h
        pad_config[1][1] = self.pad_w
        pad_config = pad_config.astype(int)
        if self.pad_h > 0 or self.pad_w > 0:
            padded_img = np.pad(img.copy(), pad_config, mode='constant')
        else:
            padded_img = img.copy()
        patches = []
        for x, y, t in self.ref_coords:
            p = padded_img[x:x + self.size, y:y + self.size]
            if random_rotate and t:
                p = p.T
            patches.append(p)
        return np.array(patches)

    def unpatchify(self, patches, n_channel):
        """
        stitch patches back into one
        :param patches: array of patches
        :param n_channel: number of channels, for instance, for a rgb image n_channel should be 3
        :return:
        """
        canvas = np.zeros(list(self.shape) + [n_channel])
        canvas_counter = np.zeros(self.shape)
        for i, p in enumerate(patches):
            x, y = self.ref_coords[i][0], self.ref_coords[i][1]
            canvas[x:x + self.size, y:y + self.size] += p
            canvas_counter[x:x + self.size, y:y + self.size] += 1
        mean_canvas = canvas / canvas_counter[:, :, np.newaxis]
        return mean_canvas[:self._shape[0], :self._shape[1]]


class MakeMask:

    def __init__(self, model,
                 description=None,
                 n_classes=1,
                 n_channels=1,
                 threshold=[0.5],
                 verbose=True,
                 **kwargs):
        """
        This function allows the users to conviniently use pre-implemented or separately trained models to do
        single- or multi-class mask prediction (semantic segment)
        """
        self.model = model
        self.description = description
        self.verbose = bool(verbose)
        self.threshold = threshold
        self.n_classes = n_classes
        self.n_channels = n_channels
        self.kwargs = kwargs

    def predict(self, targets):
        """
        targets
        """
        if not isinstance(targets, list):
            targets = [targets]

        if self.verbose:
            if self.description is not None:
                print("Here's some extra information the user provided: {}.".format(self.description))

        # mask predictions are exported as list rather than np.array, considering instances where target images have different shapes
        predictions = []

        # check target channels
        for i, t in enumerate(targets):
            t = dimension_scaler(t)
            shape = t.shape
            expected_shape = tuple(list(shape[:-1]) + [self.n_channels])
            if shape[-1] != self.n_channels:
                error_msg = 'Expected image dimension is {} but {} was given.'.format(expected_shape, shape)
                raise ValueError(error_msg)
            probabilities = self.model.predict(t, **self.kwargs)
            binary = probabilities
            predictions.append(binary)
        return predictions


class BinarizeLegend:

    def __init__(self, method=5, **kwargs):
        self.method = method
        self.kwargs = kwargs

    def predict(self, target):
        target = np.squeeze(target)
        target = target.astype(float)
        target /= target.mean()
        return _make_mask(target, method=self.method, **self.kwargs)





import yaml, pytest, os

__all__ = ['read_yaml','load_config','default_config']

def load_config(configfile = None):
    try:
        return read_yaml(configfile)
    except:
        return default_config.copy()

def read_yaml(file_path):
    with open(file_path, "r") as f:
        return yaml.safe_load(f)



default_config = {
    'AUTHOR': 'jz-rolling',
    'SYSTEM': {
        'VERBOSE': False,
    },
    'EMAIL': 'jzrolling@outlook.com',
    'FEATURE': {
        'SIGMAS': [0.3, 0.8, 1.5, 3, 5],
        'NUM_WORKERS': 12,
        'ROG': True,
        'RIDGE': True,
        'SOBEL': True,
        'SHAPEINDEX': True,
        'SELEM': [1, 'square', 5]
    },
    'IMAGE': {
        'EDGE': 0.2,
        'ROG_SIGMAS': [0.5, 10],
        'DRIFT_CORRECTION':{
            'CORRECT_DRIFT':True,
            'REFERENCE':'default',
            'INVERT':False,
            'MAX_DRIFT_PIX':5,
        }
    },
    'SEGMENTATION': {
        'SHAPEINDEX': {
            'USE_SHAPEINDEX':False,
            'SHAPEINDEX_SIGMA': 1,
            'SHAPEINDEX_LOW': 25,
            'SHAPEINDEX_HIGH': 85
        },
        'MIN_SEED_SIZE': 50,
        'MAX_HOLE_SIZE': 100,
        'MASK': {'METHOD': 1}}}




from ..utils import *
from ..segment import *
from ..plot import *
import pandas as pd
import PIL,io
from ..classify import run_particle_labeler
from matplotlib import pyplot as plt
from skimage import morphology, draw
from typing import Optional, Dict, Union, Any


__all__ = ['Patch']


class Patch:

    def __init__(self,
                 image_dict: Optional[Dict[str, Any]] = None,
                 configfile: Optional[str] = None,
                 image_id: int = 0,
                 ref_channel: Union[int, str] = -1,
                 store_backup: bool = False) -> None:
        """
        Initialize a Patch object.
        # make sure that the number of images match the number of channels provided.

        :param image_dict: Dictionary containing image data.
        :param configfile: Path to the configuration file.
        :param image_id: Text ID of the image.
        :param ref_channel: Specify the name/index of the reference channel for the image.
        :param store_backup: Flag to store unprocessed data.
        """

        self.id = image_id
        self.data = {}
        self._data = {}
        self.particles = {}
        self.cache = PatchCache()
        self.seed = None
        self.labeled_seeds = None
        self.labeled_mask = None
        self.config = load_config(configfile)
        self.mask = None
        self.prob_mask = None
        self.pixel_microns = 0.065
        self.shape = (2048, 2048)
        self.offset = (0,0)
        self.channels = []
        self.ref_channel = ''
        self.bbox = ()
        self.store_backup = store_backup
        self.regionprops = pd.DataFrame()
        self.boundaries = None
        self.wd = './'
        if isinstance(image_dict,dict):
            self.load_data(image_dict,
                           ref_channel=ref_channel)

    def load_data(self,
                  image_dict: Dict[str, np.array],
                  image_id: int=0,
                  ref_channel: Union[int,str]=-1):
        """
        Load image data into the Patch object.

        :param image_dict: Dictionary containing compiled image data: {'Channel_name':2D numpy array}.
        :param image_id: ID of the image.
        :param ref_channel: Reference channel for the image.
        """
        self.id = image_id
        self.data = image_dict
        self.channels = list(image_dict.keys())
        if isinstance(ref_channel, int):
            self.ref_channel = self.channels[ref_channel]
        elif isinstance(ref_channel, str):
            self.ref_channel = ref_channel
        else:
            raise ValueError('reference channel should be integer or string.')
        self.shape = self.data[self.ref_channel].shape
        if self.store_backup:
            self._data = self.data.copy()

    def revert(self) -> None:
        """
        Revert the image data to its original/unprocessed state.
        """
        if self.store_backup:
            self.data = self._data.copy()
        else:
            print('No raw data found. Set "store_backup" to True to backup raw data.')

    def load_seed(self,
                  seed: np.array) -> None:
        """
        Load seed image into the Patch object (to guide watershed segmentation).

        :param seed: Seed image.
        """
        if seed.shape != self.shape:
            raise ValueError("The shapes of seed image and target images don't match!")
        self.seed = seed

    def load_mask(self,
                  mask: np.array) -> None:
        """
        Load a precomputed mask into the Patch object.

        :param mask: Precomputed mask image.
        """
        if mask.shape != self.shape:
            raise ValueError("The shapes of mask image and target images don't match!")
        self.mask = mask

    def load_labeled_mask(self,
                          labeled_mask: np.array) -> None:
        """
        Load a precomputed labeled mask into the Patch object.

        :param labeled_mask: Labeled mask image.
        """
        if labeled_mask.shape != self.shape:
            raise ValueError("The shapes of labeled mask and target images don't match!")
        self.labeled_mask = labeled_mask

    def batch_process(self) -> None:
        return None

    def preprocess_images(self,
                          func=None,
                          channels = 'all', **kwargs) -> None:
        """
        Preprocess images based on a provided function and channels.

        :param func: Function to preprocess images.
        :param channels: Channels to apply the function on.
        :param kwargs: Additional keyword arguments for the function.
        """
        target_channels = []
        if isinstance(channels,str):
            if channels in ['ALL','All','all','-']:
                target_channels=self.channels
            elif channels in ['FL','fl','Fluorescence','fluorescence','F','f']:
                target_channels = [c for c in self.channels if c != self.ref_channel]
            elif channels in ['Ref','ref','Reference','reference','R','r']:
                target_channels = [self.ref_channel]
            elif channels in self.channels:
                target_channels = [channels]
            else:
                raise ValueError('Invalid channel variable "{}".'.format(channels))
        elif isinstance(channels,list):
            for c in channels:
                if c in self.channels:
                    target_channels.append(c)
                else:
                    raise ValueError('Channel "{}" not found.'.format(c))
        try:
            for c in target_channels:
                self.data[c] = func(self.data[c],**kwargs)
        except:
            print('Failed in applying function {} on these channels: {}'.format(func._name_, target_channels))

    def crop_edge(self,edge_fraction=None) -> None:
        """
        crop edge pixels to suppress edge aberrations.
        :param edge_fraction: proportion of edge pixels being cropped from each side, should be between 0 and 0.4.
        """
        if isinstance(edge_fraction,float):
            self.config['IMAGE']['EDGE'] = edge_fraction
        else:
            edge_fraction = self.config['IMAGE']['EDGE']
        if 0 <= edge_fraction < 0.4:
            crop_width = int(edge_fraction * self.shape[1])
            crop_height = int(edge_fraction * self.shape[0])
            w1, w2 = crop_width, self.shape[1] - crop_width
            h1, h2 = crop_height, self.shape[0] - crop_height
            self.shape = (h2 - h1, w2 - w1)
            self.offset = (crop_height, crop_width)
        else:
            raise ValueError('Edge fraction should be no higher than 0.4 (40% from each side)!')

        # crope edge
        for channel, data in self.data.items():
            self.data[channel] = self.data[channel][h1:h2, w1:w2]

    def correct_xy_drift(self,
                         reference_channel=None,
                         invert_ref=None,
                         max_drift=None) -> None:
        """
        Planar drift correction.
        :param reference_channel: specify reference channel, uses Patch.ref_channel by default.
        :param invert_ref: if invert reference channel data, set to True if reference channel has a light background, e.g. phase contrast.
        :param max_drift: maximum drift (in pixels) to correct for.
        """
        if reference_channel is None:
            reference_channel = self.config['IMAGE']['DRIFT_CORRECTION']['REFERENCE']
            reference_image = self.get_ref_image()
        elif isinstance(reference_channel,str):
            if reference_channel == 'default':
                reference_channel = self.ref_channel
                reference_image = self.get_ref_image()
            elif reference_channel in list(self.data.keys()):
                reference_image = self.data[reference_channel]
                self.config['IMAGE']['DRIFT_CORRECTION']['REFERENCE']=reference_channel
        else:
            raise ValueError('Channel {} not found!'.format(reference_channel))

        if invert_ref is None:
            invert_ref = self.config['IMAGE']['DRIFT_CORRECTION']['INVERT']
        elif isinstance(invert_ref,bool):
            self.config['IMAGE']['DRIFT_CORRECTION']['INVERT']=bool(invert_ref)
        else:
            raise ValueError('The variable "invert_ref" should be boolean, not {}!'.format(type(invert_ref)))

        if max_drift is None:
            max_drift = self.config['IMAGE']['DRIFT_CORRECTION']['MAX_DRIFT_PIX']
        elif isinstance(max_drift,(int,float)):
            self.config['IMAGE']['DRIFT_CORRECTION']['MAX_DRIFT_PIX']=bool(max_drift)
        else:
            raise ValueError('The variable "max_drift" should be float or int, not {}!'.format(type(max_drift)))

        # invert image if necessary
        if invert_ref:
            reference_image = 100 + reference_image.max() - reference_image
        # correct xy drift by phase cross correlation
        for channel, data in self.data.items():
            if channel != reference_channel:
                shift = drift_detection(reference_image,data,upsample_factor=10)
                if max(np.abs(shift)) <= max_drift:
                    offset_image = drift_correction(data, shift)
                    self.data[channel] = offset_image

    def generate_mask(self,
                      method=1,
                      min_particle_size=50,
                      min_hole_size=50,
                      **kwargs) -> None:
        """
        Generate a binary mask for the image.

        :param method: Numeric index or name of a pre-defined binarization method
        :param min_particle_size: Minimum size for particles in the mask.
        :param min_hole_size: Minimum size for holes in the mask.
        :param kwargs: Additional keyword arguments for the method.
        """

        mask = BinarizeLegend(method, **kwargs).predict(self.get_ref_image())
        mask = morphology.remove_small_holes(mask,min_hole_size)
        mask = morphology.remove_small_objects(mask,min_particle_size)
        self.mask = mask

    def label_mask(self) -> None:
        """
        Label the binary mask.
        """
        self.labeled_mask = measure.label(self.mask)

    def locate_particles(self,
                         use_intensity=True,
                         cache_mask = False) -> None:
        """
        Locate particles in the image.

        :param use_intensity: Use intensity data for locating particles.
        :param cache_mask: Cache the binary mask for future use.
        """
        if self.labeled_mask.max() > 0:
            if use_intensity:
                self.regionprops = get_regionprop(self.labeled_mask, intensity_image=self.get_ref_image())
            else:
                self.regionprops = get_regionprop(self.labeled_mask, intensity_image=None)
            self.regionprops.set_index('$label', inplace=True)
            self.regionprops['$image_id'] = [self.id] * len(self.regionprops)
            self.regionprops[
                ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2', '$touching_edge']] = optimize_bbox_batch(
                                                                                    self.shape, self.regionprops)
            self.regionprops['$include'] = 1
            self.regionprops['$midlines'] = [[]] * len(self.regionprops)
            self.regionprops['$outline'] = [[]] * len(self.regionprops)
            self.regionprops['$refined_outline'] = [[]] * len(self.regionprops)
            masks = []
            for l in self.regionprops.index:
                x1, y1, x2, y2 = self.regionprops.loc[l, ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2']]
                masks.append(self.labeled_mask[x1:x2, y1:y2] == l)
            if cache_mask:
                self.regionprops['$mask'] = masks

    def get_particle_data(self,particle_id):
        """
        return particle data by ID
        :param particle_id: user provided unqiue particle ID
        :return: particle data dict
        """
        if len(self.regionprops)==0:
            self.locate_particles()
        if particle_id not in self.regionprops.index:
            raise ValueError('Particle with label {} not found!'.format(particle_id))

        particle_dict = dict(self.regionprops.loc[particle_id])
        x1,y1,x2,y2 = [particle_dict[x] for x in ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2']]
        for c in self.channels:
            particle_dict[c] = self.data[c][x1:x2,y1:y2]
        if '$mask' not in particle_dict:
            particle_dict['$mask'] = (self.labeled_mask==particle_id)[x1:x2,y1:y2]
        return particle_dict

    def get_particle_mask(self,particle_id,cropped=True):
        """
        retrieve particle mask, cropped or not
        :param particle_id: user provided unqiue particle ID
        :param cropped: whether to crop the mask by optimized bbox
        :return: particle binary mask, cropped or not
        """
        if len(self.regionprops)==0:
            self.locate_particles()
        if particle_id not in self.regionprops.index:
            raise ValueError('Particle with label {} not found!'.format(particle_id))
        if not cropped:
            return (self.labeled_mask==particle_id)**1
        else:
            x1, y1, x2, y2 = self.regionprops.loc[particle_id,['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2']].values
            return (self.labeled_mask==particle_id)[x1:x2,y1:y2]

    def filter_particles(self,
                         filter_dict: Dict[str, Any]):
        """
        Filters particles in the region properties based on given criteria.

        This function performs two tasks:
        1. It removes particles that touch the edges of the image.
        2. It filters particles based on the criteria specified in the filter_dict.
           For keys starting with '$', particles are filtered based on exact value matching.
           For other keys, particles are filtered based on a value range.

        The results are stored in the '$include' column of the regionprops DataFrame,
        where particles that pass the filter are marked with a 1 and others with a 0.

        :param filter_dict: Dictionary specifying filtering criteria.
                            The keys should be feature names (possibly starting with '$')
                            and the values should be either:
                            a) An exact value for features with '$' key prefix.
                            b) A tuple (min_value, max_value) specifying a range for other features.

        :return: None. This method modifies the internal regionprops DataFrame in-place.
        """
        _accept = [self.regionprops['$touching_edge'].values == 0]
        # filter by range
        for k, r in filter_dict.items():
            if '$' in k:
                _accept.append(np.array([self.regionprops[k].values == r]))
            else:
                vmin, vmax = r
                if k in self.regionprops:
                    v = self.regionprops[k].values
                    _accept.append((v > vmin) * (v < vmax))
                else:
                    print('feature ${}$ not found!'.format(k))
        self.regionprops['$include'] = (np.sum(_accept, axis=0) == len(_accept)) * 1

    def _get_intensity_stats(self, channel):
        """
        Add basic intensity features to the regionprop table
        :params channel: specify channel name
        """
        intensity_stats = []
        if channel not in self.channels:
            raise ValueError('Channel {} not found!'.format(channel))

        channel_data = self.get_channel_data(channel)
        for coords in self.regionprops['$coords'].values:
            particle_data = channel_data[coords[:, 0], coords[:, 1]]
            _cols, particle_stat = basic_intensity_features(particle_data)
            intensity_stats.append(particle_stat)
        columns = ['{}_{}'.format(channel, c) for c in _cols]
        self.regionprops[columns] = intensity_stats

    def get_intensity_stats(self):
        """
        get particle intensity stats of all channels
        """
        for c in self.channels:
            self._get_intensity_stats(c)

    def render_image_features(self, model='default', mode='default'):
        if isinstance(model, str):
            if (model == 'default') & (len(self.cache.feature_columns) == 0):
                self._image2pixelfeatures(mode=mode)
        else:
            model.predict(self)
        return None

    def annot2feature(self):
        use_Gaussian = bool(int(self.config['classification_pixel']['use_Gaussian']))
        pixel_feature_selem = generic.config2selem(str(self.config['classification_pixel']['pixel_feature_selem']))
        self.cache.extract_local_features(use_Gaussian=use_Gaussian, selem=pixel_feature_selem)

    def mask2feature(self, ):
        coords = np.where(self.mask > 0)
        self.cache.pixel_data['xcoord'] = coords[0]
        self.cache.pixel_data['ycoord'] = coords[1]
        self.annot2feature()

    def pixel_data(self):
        return self.cache.pixel_data

    def get_ref_image(self):
        return self.data[self.ref_channel]

    def get_channel_data(self, channel):
        if isinstance(channel,str):
            if channel in self.channels:
                return self.data[channel]
            else:
                raise ValueError('Invalid channel name:{}'.format(channel))
        elif isinstance(channel,int):
            if channel<len(self.channels) and channel >=-1:
                return self.data[self.channels[channel]]
            else:
                raise ValueError('Invalid numeric channel index:{}'.format(channel))
        else:
            raise ValueError('Invalid channel format: {}. Can only be a string or integer'.format(type(channel)))

    def find_outline(self,
                     level=0.1,
                     dilation=False,
                     erosion=False,
                     calculate_curvature=True,
                     angularity_window=15,
                     dark_background=False,
                     sigma=1,
                     approximate=False,
                     tolerance=0.3,
                     interp_distance=1,
                     min_segment_count=3,
                     update_mask=True):

        if len(self.regionprops) > 0:
            bending_stat = []
            masks = []
            outline_list = []
            if '$opt-x1' not in self.regionprops:
                self.regionprops[
                    ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2', '$touching_edge']] = corrections.optimize_bbox_batch(
                    self.shape,
                    self.regionprops)
            for label in self.regionprops.index:
                x1, y1, x2, y2, touching_edge = self.regionprops.loc[label][
                    ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2', '$touching_edge']].values.astype(int)
                mask = (self.labeled_mask[x1:x2, y1:y2] == label).astype(int)
                if not touching_edge:
                    data = self.get_ref_image()[x1:x2, y1:y2]
                    outline = find_contour_marching_squares(mask, data,
                                                            level=level,
                                                            dilation=dilation,
                                                            erosion=erosion,
                                                            dark_background=dark_background,
                                                            sigma=sigma,
                                                            approximate=approximate,
                                                            tolerance=tolerance,
                                                            interp_distance=interp_distance,
                                                            min_segment_count=min_segment_count)
                    if calculate_curvature:
                        bending = -bend_angle_closed(outline, window=angularity_window)
                        bending_stat.append([bending.min(), bending.max(), bending.mean(),
                                             np.percentile(bending, 25), np.percentile(bending, 75)])
                    outline_list.append(outline)
                    if update_mask:
                        mask = draw.polygon2mask(mask.shape, outline).astype(int)
                else:
                    outline_list.append(np.array([]))
                    if calculate_curvature:
                        bending_stat.append([np.nan] * 5)
                masks.append(mask)
            self.regionprops['$outline'] = outline_list
            self.regionprops['$mask'] = masks
            if calculate_curvature:
                self.regionprops[['min_negative_curvature',
                                  'max_positive_curvature',
                                  'mean_curvature', 'Q1_curvature', 'Q3_curvature']] = bending_stat

    def extract_midlines(self,
                         min_branch_length=0.2,
                         pole_length=0.25,
                         method='zhang',
                         filtered_only=True):
        #
        midline_list = []
        cell_ids = self.regionprops.index
        for idx in cell_ids:
            cell = self.get_particle_data(idx)
            cent_x,cent_y = cell['$centroid-0']-cell['$opt-x1'],cell['$centroid-1']-cell['$opt-y1']
            mask = cell['$mask']
            mask_coords = np.array(np.where(mask>0)).T
            outline = cell['$outline']
            orientation = cell['orientation']
            solidity = cell['solidity']

            if not cell['$include'] and filtered_only:
                midlines = []
            elif len(outline) == 0:
                midlines = []
            # if cell is convex, use fast mode
            else:
                try:
                    if solidity > 0.8 and np.array([cent_x,cent_y]).astype(int) in mask_coords:
                        if len(outline) > 100:
                            upsampling_factor=1
                        else:
                            upsampling_factor=2
                        midlines = fast_midline(mask,outline,
                                                orientation=orientation,
                                                centroid=np.array([cent_x,cent_y]),
                                                pixel_microns=self.pixel_microns,
                                                pole_length=pole_length,
                                                upsampling_factor=upsampling_factor)
                    else:
                        midlines = extract_midline(mask, outline,
                                                    pixel_microns=self.pixel_microns,
                                                    min_branch_length=min_branch_length,
                                                    method=method)
                except:
                    midlines = []
            midline_list.append(midlines)
        self.regionprops['$midlines'] = midline_list

    def plot(self,
             cell_ids=[],
             figsize=(5, 5),
             outline_prop={},
             midline_prop={},
             cell_plot=False,
             channel=-1):
        print('Rendering cell plot(s)...')
        plot_cells(self,cell_ids,figsize,outline_prop,midline_prop,cell_plot,channel)

    def _image2pixelfeatures(self, mode='default'):
        if mode == 'default':
            sigmas = (0.5, 1.5, 5)
            use_RoG = True
            use_Ridge = True
            use_Sobel = False
            use_Gaussian = True
            use_Shapeindex = True
            num_workers = 12
        elif mode == 'use_config':
            # adopt parameters from configurations
            sigmas = np.array(self.config['classification_pixel']['sigmas'].split(',')).astype(float)
            use_RoG = bool(int(self.config['classification_pixel']['use_RoG']))
            use_Ridge = bool(int(self.config['classification_pixel']['use_Ridge']))
            use_Sobel = bool(int(self.config['classification_pixel']['use_Sobel']))
            use_Gaussian = bool(int(self.config['classification_pixel']['use_Gaussian']))
            use_Shapeindex = bool(int(self.config['classification_pixel']['use_Shapeindex']))
            num_workers = int(self.config['classification_pixel']['num_workers'])
        elif mode == 'fast':
            sigmas = (float(self.config['classification_pixel']['ridge_sigma']))
            use_RoG = True
            use_Ridge = True
            use_Sobel = False
            use_Gaussian = False
            use_Shapeindex = True
            num_workers = 12
        else:
            raise ValueError(
                'Illegal pixel feature extraction mode "{}", use "default", "fast", "use_config" instead.'.format(mode))
        # feature images
        self.cache.prepare_feature_images(self.get_ref_image(),
                                          num_workers=num_workers,
                                          sigmas=sigmas,
                                          rog=use_RoG,
                                          ridge=use_Ridge,
                                          sobel=use_Sobel,
                                          shapeindex=use_Shapeindex)

    def _seed2annot(self):
        from skimage.segmentation import watershed
        from skimage.measure import regionprops_table, label

        # label_seed
        self.labeled_seeds = label(self.seed)

        # adopt parameters from configurations
        ridge_sigma = float(self.config['classification_pixel']['ridge_sigma'])
        min_seed_fraction = float(self.config['classification_pixel']['min_seed_fraction'])
        erosion_selem = generic.config2selem(str(self.config['classification_pixel']['seeded_mask_erosion_selem']))
        dilation_selem = generic.config2selem(str(self.config['classification_pixel']['seeded_mask_dilation_selem']))

        # watershed segment
        ridge = filters.sato(self.get_ref_image(),
                             sigmas=(ridge_sigma),
                             black_ridges=False)
        labeled_mask = watershed(ridge, markers=self.labeled_seeds,
                                 mask=self.mask,
                                 watershed_line=True,
                                 compactness=1)
        rp = pd.DataFrame(regionprops_table(labeled_mask,
                                            intensity_image=(self.seed > 0) * 1,
                                            properties=['coords', 'mean_intensity']))
        if rp['mean_intensity'].max() < min_seed_fraction:
            x, y = np.vstack(rp[rp['mean_intensity'] < min_seed_fraction]['coords'].values).T
            labeled_mask[x, y] = 0
        self.labeled_mask = labeled_mask

        # extract relavent pixel coordinates
        seeded_mask = (labeled_mask > 0) * 1
        core = morphology.binary_erosion(seeded_mask, erosion_selem) * 1
        dilated = morphology.binary_dilation(seeded_mask, dilation_selem) * 1
        residual = ((self.mask - dilated) == 1) * 1
        edge = dilated - core

        # mask to coords
        core_coords = np.array(np.where(core > 0))
        edge_coords = np.array(np.where(edge > 0))
        residual_coords = np.array(np.where(residual > 0))

        # concat coords and annotations
        coords = np.hstack([core_coords, edge_coords, residual_coords])
        annot = [2] * len(core_coords[0]) + \
                [1] * len(edge_coords[0]) + \
                [0] * len(residual_coords[0])
        self.mask = np.zeros(self.mask.shape)
        self.mask[edge == 1] = 1
        self.mask[core == 1] = 2

        self.cache.pixel_data['xcoord'] = coords[0]
        self.cache.pixel_data['ycoord'] = coords[1]
        self.cache.pixel_data['annotation'] = annot

    def annotate_particles(self,
                           percentile_norm_params=(0, 100, 2000, 10000),
                           show_outline=False,
                           outline_color='r',
                           classes={1: 'single cell', 2: 'microcolony', 3: 'others'},
                           saturation=0.85):
        norm = (generic.percentile_normalize(self.get_ref_image(),
                                             percentile_norm_params[0],
                                             percentile_norm_params[1],
                                             percentile_norm_params[2],
                                             percentile_norm_params[3]) * int(saturation * 255)).astype(np.uint8)
        thumbnails = []
        if '$annotation' in self.regionprops.columns:
            annotation = self.regionprops['$annotation'].values
        else:
            annotation = np.zeros(len(self.regionprops))
        for (x0, y0, x1, y1, c) in self.regionprops[['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2', '$outline']].values:
            if show_outline:
                fig = plt.figure(figsize=(4, 4))
                plt.imshow(self.get_ref_image()[x0:x1, y0:y1], cmap='gist_gray')
                plt.plot(c[0][:, 1], c[0][:, 0], color=outline_color, lw=0.8)
                plt.axis('off')
                img_buf = io.BytesIO()
                plt.savefig(img_buf, bbox_inches='tight')
                thumbnails.append(PIL.Image.open(img_buf))
                plt.close()
            else:
                thumbnails.append(PIL.Image.fromarray(norm[x0:x1, y0:y1]))
        run_particle_labeler(thumbnails, label_dict=classes, annot=annotation)
        self.regionprops['$annotation'] = annotation

    def generate_multi_mask(self, model='default',
                              method=1,
                              channels=[], **kwargs):
        if len(channels) > len(self.channels):
            raise ValueError('Patch image had {} channels but {} channels were provided'.format(len(self.channels),
                                                                                                len(channels)))
        if len(channels) == 0:
            input_images = [self.get_ref_image()]

        elif len(channels) >= 1:
            stack = []
            for c in channels:
                if c in self.data:
                    stack.append(self.data[c])
                else:
                    raise ValueError('Patch image does not have channel "{}".'.format(c))
            input_images = [np.array(stack)]

        if isinstance(model, str):
            if model == 'default':
                model = BinarizeLegend(method=method, config=self.config)
            else:
                raise ModuleNotFoundError('Mask method {} not found!'.format(model))
        mask = MakeMask(model, **kwargs).predict(input_images)[0]
        self.mask = mask


class PatchCache:

    def __init__(self):
        self.gaussians = {}
        self.feature_images = {}
        self.feature_columns = []
        self.pixel_data = pd.DataFrame()

    def prepare_feature_images(self, img, **kwargs):
        self.gaussians, self.feature_images = multiscale_image_feature(img, **kwargs)

    def extract_local_features(self, use_Gaussian=True, **kwargs):
        if len(self.pixel_data) == 0:
            raise ValueError('Pixel coordinates not defined!')
        xcoord, ycoord = self.pixel_data[['xcoord', 'ycoord']].values.T
        feature_images = self.feature_images
        if use_Gaussian:
            for k, g in self.gaussians.items():
                feature_images['Gaussian_{}'.format(k)] = g
        local_pixel_features = local_stat(feature_images, xcoord, ycoord, **kwargs)
        self.feature_columns = local_pixel_features.columns
        self.pixel_data[self.feature_columns] = local_pixel_features.values

def get_regionprop(labeled_image, intensity_image=None):
    regionprop_properties = ['label','bbox','coords','centroid',
                             'area','convex_area','filled_area',
                             'eccentricity','solidity',
                             'major_axis_length','minor_axis_length',
                             'perimeter','equivalent_diameter',
                             'extent','orientation',
                             'inertia_tensor','moments_hu','weighted_moments_hu',
                             'mean_intensity','max_intensity','min_intensity']

    if intensity_image is not None:
        rp_table = pd.DataFrame(measure.regionprops_table(labeled_image,
                                                          intensity_image,
                                                          properties=regionprop_properties))
    else:
        rp_table = pd.DataFrame(measure.regionprops_table(labeled_image,
                                                          intensity_image=(labeled_image>0)*1,
                                                          properties=regionprop_properties))
    areas = rp_table['area'].values
    perimeter = rp_table['perimeter'].values
    rp_table['compactness'] = (perimeter**2)/(4*np.pi*areas)
    rp_table['rough_length'] = (perimeter-np.sqrt(perimeter**2-16*areas))/4
    rp_table['aspect_ratio'] = np.round(rp_table['minor_axis_length'].values/rp_table['major_axis_length'].values,3)
    rp_table.rename(columns={'bbox-0':'$bbox-0',
                             'bbox-1':'$bbox-1',
                             'bbox-2':'$bbox-2',
                             'bbox-3':'$bbox-3',
                             'coords':'$coords',
                             'label':'$label',
                             'centroid-0':'$centroid-0',
                             'centroid-1':'$centroid-1'},
                    inplace=True)
    return rp_table

    """
    def load_config(self, config):
        if 'ConfigParser' in str(type(config)):
            self.config = config
        else:
            self.config = Config(config).config
    """







import numpy as np

from ..utils import *
from ..core.patch import Patch

class Particle:

    def __init__(self,
                 bbox,
                 mask,
                 data,
                 ref_channel,
                 particle_id,
                 offset,
                 pixel_microns=0.065,
                 outline=np.array([[]])):
        self.bbox = bbox
        self.outline = outline
        self.mask = mask
        self.x,self.y = np.where(mask>0)
        self.data = data
        self.ref_channel = ref_channel
        self.pixel_microns = pixel_microns
        self.id = particle_id
        self.sobel = None
        self.branched = False
        self.discarded = False
        self.skeleton = None
        self.skeleton_coords = {}
        self.midlines = {}
        self.branch_lengths = {}
        self.width_lists = {}
        self.length = 0
        self.config = None
        self.offset = offset
        self.basic_feature_attributes = ['area', 'eccentricity', 'solidity', 'touching_edge',
                                         'convex_area', 'filled_area', 'eccentricity', 'solidity',
                                         'major_axis_length', 'minor_axis_length', 'perimeter',
                                         'equivalent_diameter', 'extent','circularity', 'aspect_ratio',
                                         'moments_hu-0', 'moments_hu-1', 'moments_hu-2', 'moments_hu-3',
                                         'moments_hu-4', 'moments_hu-5', 'moments_hu-6', 'aspect_ratio',
                                         'bending_std', 'bending_90', 'bending_10', 'bending_max',
                                         'bending_min', 'bending_skewness', 'bending_median', 'touching_edge']
        self.basic_features = []

    def _load_basic_features(self, parent_regionprop, basic_attributes):
        self.basic_feature_attributes = basic_attributes
        self.basic_features = parent_regionprop[basic_attributes].values

    def get_ref_image(self):
        return self.data[self.ref_channel]


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simplified_particle_labler.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.

import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QKeySequence
from PIL import Image, ImageQt
from PyQt5.QtGui import QImage, QPainter, QPixmap, QPalette
import sys
from functools import partial


class ParticleLabelerMain(QtWidgets.QWidget):

    def __init__(self, images=[], label_dict={}, annot=[], *args, **kwargs):
        QtWidgets.QWidget.__init__(self, *args, **kwargs)
        if len(images) > 0:
            self.images = images
        else:
            self.images = [(np.random.random((50, 50)) * 255).astype(np.uint8) for i in range(5)]
        if len(annot) == 0:
            self.annotation = np.zeros(len(self.images))
        else:
            self.annotation = annot
        self.default_labeldict = label_dict
        self.mainwindow = QtWidgets.QMainWindow()
        self.setupUi(self.mainwindow)

    def setupUi(self, MainWindow):
        self.Nlablers = 2
        self.MaxNlablers = 10
        self.labelers = []
        self.counter = np.where(self.annotation == 0)[0][0]
        self.Nimage = len(self.images)
        self.enableFastLabelMode = False
        MainWindow.setObjectName("ParticleLablerMain")
        MainWindow.resize(791, 599)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.particle_viewer_container = QtWidgets.QGroupBox(self.centralwidget)
        self.particle_viewer_container.setGeometry(QtCore.QRect(10, 10, 501, 511))
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.particle_viewer_container.setFont(font)
        self.particle_viewer_container.setObjectName("particle_viewer_container")
        self.particle_viewer = QtWidgets.QLabel(self.particle_viewer_container)
        self.particle_viewer.setBackgroundRole(QPalette.Base)
        self.particle_viewer.setGeometry(QtCore.QRect(10, 30, 481, 461))
        self.particle_viewer.setObjectName("particle_viewer")
        self.particle_scroller = QtWidgets.QScrollBar(self.particle_viewer_container)
        self.particle_scroller.setGeometry(QtCore.QRect(10, 490, 421, 20))
        self.particle_scroller.setOrientation(QtCore.Qt.Horizontal)
        self.particle_scroller.setInvertedAppearance(False)
        self.particle_scroller.setInvertedControls(False)
        self.particle_scroller.setRange(1, self.Nimage)
        self.particle_scroller.setObjectName("particle_scroller")
        self.particle_scroller.valueChanged.connect(self.scroll_action)
        self.particle_N_track = QtWidgets.QLabel(self.particle_viewer_container)
        self.particle_N_track.setGeometry(QtCore.QRect(430, 490, 60, 16))
        self.particle_N_track.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter)
        self.particle_N_track.setObjectName("particle_N_track")
        font = QtGui.QFont()
        font.setPointSize(11)
        self.particle_info = QtWidgets.QLabel(MainWindow)
        self.particle_info.setGeometry(QtCore.QRect(200, 10, 300, 30))
        self.particle_info.setAlignment(QtCore.Qt.AlignRight)
        self.particle_info.setObjectName("particle_info")

        self.labler_container = QtWidgets.QGroupBox(self.centralwidget)
        self.labler_container.setGeometry(QtCore.QRect(520, 10, 261, 511))
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        self.labler_container.setFont(font)
        self.labler_container.setObjectName("labler_container")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.labler_container)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(9, 50, 241, 451))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.hotkey = QtWidgets.QLabel(self.labler_container)
        self.hotkey.setGeometry(QtCore.QRect(0, 30, 51, 16))
        self.hotkey.setAlignment(QtCore.Qt.AlignCenter)
        self.hotkey.setObjectName("hotkey")
        self.class_name_header = QtWidgets.QLabel(self.labler_container)
        self.class_name_header.setGeometry(QtCore.QRect(100, 30, 81, 16))
        self.class_name_header.setAlignment(QtCore.Qt.AlignCenter)
        self.class_name_header.setObjectName("class_name_header")
        self.revert2default_button = QtWidgets.QPushButton(self.centralwidget)
        self.revert2default_button.setGeometry(QtCore.QRect(290, 530, 131, 32))
        self.revert2default_button.setMouseTracking(False)
        self.revert2default_button.setAutoDefault(False)
        self.revert2default_button.setObjectName("revert2default_button")
        self.finish_labeling_button = QtWidgets.QPushButton(self.centralwidget)
        self.finish_labeling_button.setGeometry(QtCore.QRect(420, 530, 91, 32))
        self.finish_labeling_button.setObjectName("finish_labeling_button")
        self.finish_labeling_button.clicked.connect(self.safeExit)

        self.add_Class_button = QtWidgets.QPushButton(self.centralwidget)
        self.add_Class_button.setGeometry(QtCore.QRect(520, 530, 60, 32))
        self.add_Class_button.setObjectName("add_Class_button")
        self.add_Class_button.clicked.connect(self.addClass)
        self.del_Class_button = QtWidgets.QPushButton(self.centralwidget)
        self.del_Class_button.setGeometry(QtCore.QRect(580, 530, 60, 32))
        self.del_Class_button.setObjectName("dell_Class_button")
        self.del_Class_button.clicked.connect(self.delClass)

        self.prev_particle_button = QtWidgets.QPushButton(self.centralwidget)
        self.prev_particle_button.setGeometry(QtCore.QRect(10, 530, 131, 32))
        self.prev_particle_button.setObjectName("prev_particle_button")
        self.prev_particle_button.clicked.connect(self.previous_image)
        self.next_particle_button = QtWidgets.QPushButton(self.centralwidget)
        self.next_particle_button.setGeometry(QtCore.QRect(140, 530, 131, 32))
        self.next_particle_button.setObjectName("next_particle_button")
        self.next_particle_button.clicked.connect(self.next_image)
        self.line = QtWidgets.QFrame(self.centralwidget)
        self.line.setGeometry(QtCore.QRect(270, 530, 20, 31))
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.line_2 = QtWidgets.QFrame(self.centralwidget)
        self.line_2.setGeometry(QtCore.QRect(510, 530, 20, 31))
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.label_mode = QtWidgets.QPushButton(self.centralwidget)
        self.label_mode.setGeometry(QtCore.QRect(640, 530, 149, 32))
        self.label_mode.setObjectName("label_mode")
        self.label_mode.clicked.connect(self.startAnnotation)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.retranslateUi(MainWindow)
        self.presetLablers(label_dict=self.default_labeldict)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        self.display_n()
        self.prev_butt_shortcut = QtWidgets.QShortcut(QKeySequence("Left"), self.prev_particle_button)
        self.prev_butt_shortcut.activated.connect(self.previous_image)
        self.next_butt_shortcut = QtWidgets.QShortcut(QKeySequence("Right"), self.next_particle_button)
        self.next_butt_shortcut.activated.connect(self.next_image)
        self.initLablerShortcut()
        self.watchFastLabel()
        self.update_scroller()

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.particle_viewer_container.setTitle(_translate("MainWindow", "Particle viewer"))
        self.particle_N_track.setText(_translate("MainWindow", "1/{}".format(self.Nimage)))
        self.labler_container.setTitle(_translate("MainWindow", "Class labeler"))
        self.hotkey.setText(_translate("MainWindow", "Hotkey"))
        self.class_name_header.setText(_translate("MainWindow", "Class name"))
        self.revert2default_button.setText(_translate("MainWindow", "Revert to default"))
        self.finish_labeling_button.setText(_translate("MainWindow", "FINISH"))
        self.add_Class_button.setText(_translate("MainWindow", "+"))
        self.del_Class_button.setText(_translate("MainWindow", "-"))
        self.prev_particle_button.setText(_translate("MainWindow", "Previous (<-)"))
        self.next_particle_button.setText(_translate("MainWindow", "Next (->)"))
        self.label_mode.setText(_translate("MainWindow", "Start annotation"))
        self.update_particle_info()

    def presetLablers(self, label_dict={}, enable_edit=False):
        _translate = QtCore.QCoreApplication.translate
        self.Nlablers = max(2, len(label_dict))
        for i in range(self.MaxNlablers):
            idx = i + 1
            key = (i + 1) % 10
            if idx > self.Nlablers:
                hide = True
            else:
                hide = False
            font = QtGui.QFont()
            font.setFamily("Arial")
            font.setPointSize(12)
            widget = QtWidgets.QWidget(self.verticalLayoutWidget)
            widget.setObjectName("class_{}_widget".format(idx))
            label = QtWidgets.QLabel(widget)
            label.setGeometry(QtCore.QRect(0, 1, 31, 20))
            label.setAlignment(QtCore.Qt.AlignCenter)
            label.setObjectName("class_{}_label".format(idx))
            name = QtWidgets.QLineEdit(widget)
            name.setGeometry(QtCore.QRect(32, 0, 201, 21))
            name.setObjectName("class_{}_name".format(idx))
            name.setFont(font)
            label.setFont(font)
            label.setText(_translate('MainWindow', str(key)))
            name.setText(_translate('MainWindow', 'Class {}'.format(idx)))
            if idx in label_dict:
                name.setText(_translate('MainWindow', label_dict[idx]))
            self.verticalLayout.addWidget(widget)
            if hide:
                name.setEnabled(False)
                name.hide()
                label.hide()
            self.labelers.append([widget, label, name])
            if enable_edit:
                self._enable_edit(True)
            else:
                self._enable_edit(False)

    def addClass(self):
        n = min(self.Nlablers, self.MaxNlablers - 1)
        widget, label, name = self.labelers[n]
        name.setEnabled(True)
        name.show()
        label.show()
        self.Nlablers = min(self.Nlablers + 1, self.MaxNlablers - 1)

    def delClass(self):
        n = max(self.Nlablers - 1, 0)
        widget, label, name = self.labelers[n]
        name.setEnabled(False)
        name.hide()
        label.hide()
        self.Nlablers = max(self.Nlablers - 1, 0)

    def startAnnotation(self):
        if self.label_mode.text() == "Start annotation":
            self._enable_edit(False)
        else:
            self._enable_edit(True)

    def _enable_edit(self, edit_on=True):
        if not edit_on:
            for widget, label, name in self.labelers:
                name.setEnabled(False)
                self.label_mode.setText('Edit class labels')
            self.add_Class_button.setEnabled(False)
            self.del_Class_button.setEnabled(False)
            self.enableFastLabelMode = True
        else:
            for widget, label, name in self.labelers:
                name.setEnabled(True)
                self.label_mode.setText('Start annotation')
            self.add_Class_button.setEnabled(True)
            self.del_Class_button.setEnabled(True)
            self.enableFastLabelMode = False

    def update_scroller(self):
        self.particle_scroller.setValue(self.counter + 1)
        self.update_particle_tracker()

    def scroll_action(self):
        self.counter = self.particle_scroller.value() - 1
        self.display_n()
        self.update_particle_tracker()

    def update_particle_tracker(self):
        _translate = QtCore.QCoreApplication.translate
        self.particle_N_track.setText(_translate("MainWindow", "{}/{}".format(self.counter + 1, self.Nimage)))

    def initLablerShortcut(self):
        self.shortcuts = []
        for i in range(len(self.labelers)):
            shortcut = str((i + 1) % 10)
            shortcut_obj = QtWidgets.QShortcut(QKeySequence(shortcut), self.next_particle_button)
            self.shortcuts.append(shortcut_obj)

    def initFastLabel(self, annot):
        if self.enableFastLabelMode and annot < self.Nlablers:
            self.annotation[self.counter] = annot + 1
            self.next_image()

    def watchFastLabel(self):
        for i, v in enumerate(self.shortcuts):
            func = partial(self.initFastLabel, i)
            v.activated.connect(func)

    def _display_image(self, pil_img):
        shape = pil_img.size
        if shape[1] > shape[0]:
            pil_img = pil_img.transpose(Image.ROTATE_90)
            shape = pil_img.size
        hcenter = 261
        wcenter = 251
        max_half_height = 230
        max_half_width = 240
        relative_height = shape[1] / shape[0]
        trimmed_half_height = int(round(relative_height * max_half_height))

        new_window_width = 2 * trimmed_half_height

        self.particle_viewer.setGeometry(QtCore.QRect(10, hcenter - trimmed_half_height, 481, new_window_width))
        img = ImageQt.ImageQt(pil_img)
        self.particle_viewer.setPixmap(QPixmap.fromImage(img))
        self.particle_viewer.setScaledContents(True)

    def update_particle_info(self):
        _translate = QtCore.QCoreApplication.translate
        n = self.counter + 1
        class_id = int(self.annotation[self.counter])
        if class_id == 0:
            info = "Particle {} hasn't been annotated.".format(n)
        else:
            class_label = self.labelers[class_id - 1][2].text()
            info = "Particle {} is annotated as class: {} ({})".format(n, class_id, class_label)
        self.particle_info.setText(_translate("MainWindow", info))

    def display_n(self):
        self._display_image(self.images[self.counter])
        self.update_particle_info()

    def previous_image(self):
        self.counter = max(0, self.counter - 1)
        self.update_scroller()
        self.display_n()

    def next_image(self):
        self.counter = min(self.Nimage - 1, self.counter + 1)
        self.update_scroller()
        self.display_n()

    def safeExit(self):
        """exit the application gently so Spyder IDE will not hang"""
        self.mainwindow.deleteLater()
        self.mainwindow.close()
        self.mainwindow.destroy()


def run_particle_labeler(images=[(np.random.random((100, 50)) * 255).astype(np.uint8) for i in range(100)],
                         label_dict={1: 'single cell', 2: 'microcolony', 3: 'others'},
                         annot=[]):
    # if __name__ == '__main__':
    app = QtCore.QCoreApplication.instance()
    app = None
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
        app.aboutToQuit.connect(app.deleteLater)
        main = ParticleLabelerMain(images=images, label_dict=label_dict, annot=annot)
        main.mainwindow.show()
        app.exec_()


from .generic import *
from .linalg import *

"""
============================= Description =============================
contour/outline manipulation
"""

def find_contour_marching_squares(binary_mask,
                                  image = None,
                                  level=0.1,
                                  dilation=False,
                                  erosion=True,
                                  dark_background=False,
                                  sigma=1,
                                  approximate=False,
                                  tolerance=0.3,
                                  interp_distance=1,
                                  min_segment_count=3,
                                  ):
    """
    finds the outermost contour of a binary mask using the marching-squares algorithm.
    :param binary_mask: binary mask of the object
    :param image: optional bright-field image to use as reference
    :param level: threshold level for contour detection
    :param dilation: boolean flag to perform binary dilation on the mask
    :param erosion: boolean flag to perform binary erosion on the mask
    :param dark_background: boolean flag to indicate if the image has a dark background
    :param sigma: sigma value for Gaussian smoothing of the image
    :return: outermost contour of the object
    """
    from skimage import morphology, filters, measure

    # smoothen mask
    binary_mask = smooth_binary_mask(binary_mask.astype(int),sigma=1)

    # use local threshold to suppress shadow signals (usually from cell debris)
    dil_mask = morphology.binary_dilation(binary_mask,morphology.disk(2))
    th = filters.threshold_otsu(image[dil_mask == 1].flatten())
    binary_mask = dil_mask * (image < th)

    # erode or dilate
    if erosion:
        binary_mask = morphology.binary_erosion(binary_mask).astype(int)
    if dilation:
        binary_mask = morphology.binary_dilation(binary_mask).copy()

    # use the bright-field image as reference if available:
    if image is not None:
        if not identical_shapes([image, binary_mask]):
            raise ValueError('Input images are of different sizes!')
        if not dark_background:
            image = gaussian_smooth(invert_normalize(image),sigma=sigma)
        else:
            image = normalize_image(gaussian_smooth(image,sigma=sigma),min_perc=0,max_perc=0.995)
        intensity_mask = binary_mask*image
    else:
        intensity_mask = binary_mask
    contour = measure.find_contours(intensity_mask, level=level)[0]
    if approximate:
        contour = simplify_polygon(contour,
                                   tolerance=tolerance,
                                   interp_distance=interp_distance,
                                   min_segment_count=min_segment_count)
        contour = spline_approximation(contour,n=len(contour),smooth_factor=1,closed=True)
    return contour


@jit(nopython=True, cache=True)
def estimate_outline_length(closed_coords):
    max_L = 0
    for i in range(len(closed_coords)):
        for j in range(i+1,len(closed_coords)):
            x1,y1 = closed_coords[i]
            x2,y2 = closed_coords[j]
            dist =  np.sqrt((x2-x1)**2+(y2-y1)**2)
            if dist > max_L:
                max_L = dist
    return max_L
